# -*- coding: utf-8 -*-

import pytest
import time
from .test_base_class import TestBaseClass

import aerospike


@pytest.mark.usefixtures("connection_config")
class TestChangePassword(object):

    pytestmark = pytest.mark.skipif(
        not TestBaseClass.auth_in_use(), reason="No user specified, may be not secured cluster."
    )

    def setup_method(self, method):
        """
        Setup method
        """
        config = TestBaseClass.get_connection_config()
        self.client = aerospike.client(config).connect(config["user"], config["password"])

        try:
            self.client.admin_create_user("testchangepassworduser", "aerospike", ["read"], {})
            time.sleep(2)
        except aerospike.exception.UserExistsError:
            pass  # we are good, no such role exists
        self.delete_users = []

    def teardown_method(self, method):
        """
        Teardown method
        """

        try:
            self.client.admin_drop_user("testchangepassworduser")
        except Exception:
            pass

        self.client.close()

    def test_change_password_without_any_parameters(self):

        with pytest.raises(TypeError):
            self.client.admin_change_password()

    # NOTE: This will fail if auth_mode is PKI_AUTH (3).
    @pytest.mark.xfail(reason="Might fail depending on auth_mode.")
    def test_change_password_with_proper_parameters(self):

        user = "testchangepassworduser"
        config = self.connection_config
        self.clientreaduser = aerospike.client(config).connect(user, "aerospike")

        password = "newpassword"

        status = self.clientreaduser.admin_change_password(user, password)

        assert status == 0

        time.sleep(2)
        config = self.connection_config

        # Assert that connecting to the server with the old password fails
        with pytest.raises((aerospike.exception.InvalidPassword, aerospike.exception.InvalidCredential)):
            self.clientreaduserwrong = aerospike.client(config).connect(user, "aerospike")

        self.clientreaduserright = aerospike.client(config).connect(user, "newpassword")

        assert self.clientreaduserright is not None

        self.clientreaduserright.close()
        self.clientreaduser.close()

    def test_change_password_with_invalid_timeout_policy_value(self):

        policy = {"timeout": 0.1}
        user = "testchangepassworduser"
        password = "newpassword"

        try:
            self.client.admin_change_password(user, password, policy)

        except aerospike.exception.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "timeout is invalid"

    # NOTE: This will fail if auth_mode is PKI_AUTH (3).
    @pytest.mark.xfail(reason="Might fail depending on auth_mode.")
    def test_change_password_with_proper_timeout_policy_value(self):

        user = "testchangepassworduser"
        config = self.connection_config
        self.clientreaduser = aerospike.client(config).connect(user, "aerospike")

        policy = {"timeout": 1000}
        password = "newpassword"

        status = self.clientreaduser.admin_change_password(user, password, policy)

        assert status == 0
        time.sleep(2)
        config = self.connection_config

        with pytest.raises((aerospike.exception.InvalidPassword, aerospike.exception.InvalidCredential)):
            self.clientreaduserwrong = aerospike.client(config).connect(user, "aerospike")

        self.clientreaduserright = aerospike.client(config).connect(user, "newpassword")

        assert self.clientreaduserright is not None

        self.clientreaduserright.close()
        self.clientreaduser.close()

    def test_change_password_with_none_username(self):

        policy = {}
        user = None
        password = "newpassword"

        try:
            self.client.admin_change_password(user, password, policy)

        except aerospike.exception.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Username should be a string"

    def test_change_password_with_none_password(self):

        policy = {}
        user = "testchangepassworduser"
        password = None

        try:
            self.client.admin_change_password(user, password, policy)

        except aerospike.exception.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Password should be a string"

    def test_change_password_with_non_existent_user(self):

        policy = {}
        user = "readwriteuser"
        password = "newpassword"

        try:
            self.client.admin_change_password(user, password, policy)

        except aerospike.exception.InvalidUser as exception:
            assert exception.code == 60
            assert exception.msg == "AEROSPIKE_INVALID_USER"

    def test_change_password_with_too_long_password(self):

        user = "testchangepassworduser"
        config = self.connection_config
        self.clientreaduser = aerospike.client(config).connect(user, "aerospike")

        policy = {"timeout": 100}
        password = "password" * 1000

        with pytest.raises(aerospike.exception.ClientError):
            self.clientreaduser.admin_change_password(user, password, policy)

        self.clientreaduser.close()



"""
Tests to demonstrate correct calculations of key digests
"""
import pytest
import aerospike
from binascii import b2a_hex


NAMESPACE = "namespace"
SET = "set"

# Tests for integer keys


def convert_bytestr_to_str(byte_str):
    converted = byte_str
    # This is true in Python 3
    if bytes is not str:
        converted = str(converted, "utf-8")
    return converted


@pytest.mark.parametrize(
    "pk, expected",
    (
        (-1, "22116d253745e29fc63fdf760b6e26f7e197e01d"),
        (0, "93d943aae37b017ad7e011b0c1d2e2143c2fb37d"),
        (1, "82d7213b469812947c109a6d341e3b5b1dedec1f"),
        # 8it boundaries
        (-128, "93191e549f8f3548d7e2cfc958ddc8c65bcbe4c6"),
        (127, "a58f7d98bf60e10fe369c82030b1c9dee053def9"),
        (255, "5a7dd3ea237c30c8735b051524e66fd401a10f6a"),
        # 16 bit boundaries
        (-(2**15), "7f41e9dd1f3fe3694be0430e04c8bfc7d51ec2af"),
        (2**15 - 1, "309fc9c2619c4f65ff7f4cd82085c3ee7a31fc7c"),
        (2**16 - 1, "3f0dd44352749a9fd5b7ec44213441ef54c46d57"),
        # 32 bit boundaries
        (-(2**31), "d635a867b755f8f54cdc6275e6fb437df82a728c"),
        (2**31 - 1, "fa8c47b8b898af1bbcb20af0d729ca68359a2645"),
        (2**32 - 1, "2cdf52bf5641027042b9cf9a499e509a58b330e2"),
        # 64 bit boundaries
        (-(2**63), "7185c2a47fb02c996daed26b4e01b83240aee9d4"),
        (2**63 - 1, "1698328974afa62c8e069860c1516f780d63dbb8"),
    ),
)
def test_validate_digest_ints(pk, expected):
    digest = aerospike.calc_digest(NAMESPACE, SET, pk)
    hex_digest = b2a_hex(digest)
    assert convert_bytestr_to_str(hex_digest) == expected


"""
These strings are tested in the other clients' digest tests,
but the max binname allowed is 14 characters (15 bytes):
https://github.com/aerospike/aerospike-server/blob/master/as/include/base/datamodel.h#L130
"""


@pytest.mark.parametrize(
    "pk, expected",
    (
        ("", "2819b1ff6e346a43b4f5f6b77a88bc3eaac22a83"),
        ("s", "607cddba7cd111745ef0a3d783d57f0e83c8f311"),
        ("a" * 10, "5979fb32a80da070ff356f7695455592272e36c2"),
        ("m" * 100, "f00ad7dbcb4bd8122d9681bca49b8c2ffd4beeed"),
        ("t" * 1000, "07ac412d4c33b8628ab147b8db244ce44ae527f8"),
        ("-" * 10000, "b42e64afbfccb05912a609179228d9249ea1c1a0"),
        ("+" * 100000, "0a3e888c20bb8958537ddd4ba835e4070bd51740"),
    ),
)
def test_validate_digest_str(pk, expected):
    digest = aerospike.calc_digest(NAMESPACE, SET, pk)
    hex_digest = b2a_hex(digest)
    assert convert_bytestr_to_str(hex_digest) == expected


"""
Some clients seem to allow computation with a 0 length bytearray,
Python does not, so we omit it
"""


@pytest.mark.parametrize(
    "pk, expected",
    (
        ("s", "ca2d96dc9a184d15a7fa2927565e844e9254e001"),
        ("a" * 10, "d10982327b2b04c7360579f252e164a75f83cd99"),
        ("m" * 100, "475786aa4ee664532a7d1ea69cb02e4695fcdeed"),
        ("t" * 1000, "5a32b507518a49bf47fdaa3deca53803f5b2e8c3"),
        ("-" * 10000, "ed65c63f7a1f8c6697eb3894b6409a95461fd982"),
        ("+" * 100000, "fe19770c371774ba1a1532438d4851b8a773a9e6"),
    ),
)
def test_validate_digest_bytes(pk, expected):
    # encode the primary key as bytes
    byte_key = bytearray(pk, "utf-8")
    digest = aerospike.calc_digest(NAMESPACE, SET, byte_key)
    hex_digest = b2a_hex(digest)
    assert convert_bytestr_to_str(hex_digest) == expected


def test_calculating_a_digest_with_no_set():
    expected = "a443f05d05d962202b59abb402afae1737dbf66a"
    digest = aerospike.calc_digest(NAMESPACE, "", 1)
    hex_digest = b2a_hex(digest)
    assert convert_bytestr_to_str(hex_digest) == expected



# -*- coding: utf-8 -*-

import pytest
from .test_base_class import TestBaseClass

import aerospike


class TestCalcDigest(object):
    @pytest.mark.parametrize(
        "ns, set, key",
        [
            ("test", "demo", 1),  # integer key
            ("test", "demo", "get_key_digest"),  # string key
            ("test", "demo", "get_key_digest"),  # unicode key
            ("test", "set", "get_key_digest"),  # unicode set
            ("test", "demo", bytearray("askluy3oijs", "utf-8")),  # bytearray key
        ],
    )
    def test_pos_calc_digest_with_key(self, ns, set, key):
        """
        Invoke calc_digest() with key
        """
        digest = aerospike.calc_digest(ns, set, key)

        assert isinstance(digest, bytearray)

    def test_pos_calc_digest_on_inserted_key(self):
        """
        Invoke calc_digest() on inserted key
        """
        client = TestBaseClass.get_new_connection()

        key = ("test", "demo", "get_digest_key")
        client.put(key, {"bin1": 1})
        key, meta, bins = client.get(key)

        assert key[3] == aerospike.calc_digest("test", "demo", "get_digest_key")

        client.remove(key)
        client.close()

    # Negative Tests
    def test_neg_calc_digest_with_no_parameter(self):
        """
        Invoke calc_digest() without any mandatory parameters.
        """
        with pytest.raises(TypeError) as typeError:
            aerospike.calc_digest()

        assert "argument 'ns' (pos 1)" in str(typeError.value)

    def test_neg_calc_digest_with_only_ns(self):
        """
        Invoke calc_digest() with a ns and no other parameters
        """
        with pytest.raises(TypeError) as typeError:
            aerospike.calc_digest("test")

        assert "argument 'set' (pos 2)" in str(typeError.value)

    @pytest.mark.parametrize(
        "ns, set, key, err_msg",
        [
            (1, "demo", "get_digest_key", "Namespace should be a string"),
            # Namespace is a int
            ("test", 1, "get_digest_key", "Set should be a string or unicode"),
            # Namespace is a int
            ("test", "demo", [1, 2, 3], "Key is invalid"),  # Key is a list
            ("test", "set", {"a": 1, "b": 2}, "Key is invalid"),  # Key is a map
            ("test", "demo", None, "Key is invalid"),  # Key is None
        ],
    )
    def test_neg_calc_digest_with_invalid_parameters(self, ns, set, key, err_msg):
        """
        Invoke calc_digest() with invalid parameters
        """
        with pytest.raises(TypeError) as typeError:
            aerospike.calc_digest(ns, set, key)

        assert err_msg in str(typeError.value)



# -*- coding: utf-8 -*-

import pytest
import time
from .test_base_class import TestBaseClass
from aerospike import exception as e

import aerospike


@pytest.mark.usefixtures("connection_config")
class TestDropUser(object):

    pytestmark = pytest.mark.skipif(
        not TestBaseClass.auth_in_use(), reason="No user specified, may be not secured cluster."
    )

    def setup_method(self, method):
        """
        Setup method.
        """
        config = TestBaseClass.get_connection_config()
        TestDropUser.Me = self
        self.client = aerospike.client(config).connect(config["user"], config["password"])
        try:
            self.client.admin_drop_user("foo-test")
            time.sleep(2)
        except Exception:
            pass

    def teardown_method(self, method):
        """
        Teardoen method.
        """
        self.client.close()

    def test_drop_user_with_no_parameters(self):
        """
        Invoke drop_user() without any mandatory parameters.
        """
        with pytest.raises(TypeError) as typeError:
            self.client.admin_drop_user()

        assert "argument 'user' (pos 1)" in str(typeError.value)

    def test_drop_user_with_policy_none(self):
        """
        Invoke drop_user() with policy none
        """
        policy = None
        user = "foo-test"
        password = "foo1"
        roles = ["read", "read-write", "sys-admin"]

        status = self.client.admin_create_user(user, password, roles, policy)

        time.sleep(2)

        assert status == 0
        user_details = self.client.admin_query_user(user, policy)

        assert user_details == ["read", "read-write", "sys-admin"]

        status = self.client.admin_drop_user(user, policy)

        assert status == 0

        try:
            user_details = self.client.admin_query_user(user)

        except e.InvalidUser as exception:
            assert exception.code == 60
            assert exception.msg == "AEROSPIKE_INVALID_USER"

    def test_drop_user_with_user_none(self):
        """
        Invoke drop_user() with policy none
        """
        policy = {"timeout": 1000}
        try:
            self.client.admin_drop_user(None, policy)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Username should be a string"

    def test_drop_user_positive(self):
        """
        Invoke drop_user() with correct arguments.
        """
        policy = {"timeout": 1000}
        user = "foo-test"
        password = "foo1"
        roles = ["read", "read-write", "sys-admin"]

        status = self.client.admin_create_user(user, password, roles, policy)

        time.sleep(1)

        assert status == 0
        user_details = self.client.admin_query_user(user, policy)

        assert user_details == ["read", "read-write", "sys-admin"]
        status = self.client.admin_drop_user(user, policy)
        assert status == 0

        time.sleep(2)

        try:
            user_details = self.client.admin_query_user(user, policy)

        except e.InvalidUser as exception:
            assert exception.code == 60
            assert exception.msg == "AEROSPIKE_INVALID_USER"

    def test_drop_user_positive_without_policy(self):
        """
        Invoke drop_user() with correct arguments.
        """
        policy = {"timeout": 1000}
        user = "foo-test"
        password = "foo1"
        roles = ["read", "read-write", "sys-admin"]

        status = self.client.admin_create_user(user, password, roles, policy)

        time.sleep(1)

        assert status == 0
        user_details = self.client.admin_query_user(user, policy)

        assert user_details == ["read", "read-write", "sys-admin"]
        status = self.client.admin_drop_user(user)
        assert status == 0

        time.sleep(1)

        try:
            user_details = self.client.admin_query_user(user, policy)

        except e.InvalidUser as exception:
            assert exception.code == 60
            assert exception.msg == "AEROSPIKE_INVALID_USER"

    def test_drop_user_negative(self):
        """
        Invoke drop_user() with non-existent user.
        """
        policy = {}
        user = "foo-test"
        try:
            self.client.admin_query_user(user, policy)

        except e.InvalidUser as exception:
            assert exception.code == 60
            assert exception.msg == "AEROSPIKE_INVALID_USER"

        try:
            self.client.admin_drop_user(user)

        except e.InvalidUser as exception:
            assert exception.code == 60
            assert exception.msg == "AEROSPIKE_INVALID_USER"

    def test_drop_user_policy_incorrect(self):
        """
        Invoke drop_user() with policy incorrect
        """
        policy = {"timeout": 1000}
        user = "incorrect-policy"
        password = "foo1"
        roles = ["read", "read-write", "sys-admin"]

        status = self.client.admin_create_user(user, password, roles, policy)

        time.sleep(1)

        assert status == 0
        user_details = self.client.admin_query_user(user, policy)

        assert user_details == ["read", "read-write", "sys-admin"]
        policy = {"timeout": 0.2}
        try:
            status = self.client.admin_drop_user(user, policy)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "timeout is invalid"

        status = self.client.admin_drop_user(user)

    def test_drop_user_with_extra_argument(self):
        """
        Invoke drop_user() with extra argument.
        """
        policy = {"timeout": 1000}
        with pytest.raises(TypeError) as typeError:
            self.client.admin_drop_user("foo-test", policy, "")

        assert "admin_drop_user() takes at most 2 arguments (3 given)" in str(typeError.value)

    @pytest.mark.xfail(reason="It is no longer possible to create a user with" "a name too long")
    def test_drop_user_with_too_long_username(self):

        policy = {}
        user = "user$" * 1000
        password = "user10"
        roles = ["sys-admin"]

        try:
            self.client.admin_create_user(user, password, roles, policy)

        except e.InvalidUser as exception:
            assert exception.code == 60
            assert exception.msg == "AEROSPIKE_INVALID_USER"

        try:
            self.client.admin_drop_user(user, policy)

        except e.InvalidUser as exception:
            assert exception.code == 60
            assert exception.msg == "AEROSPIKE_INVALID_USER"

    def test_drop_user_with_special_characters_in_username(self):

        policy = {}
        user = "!#Q#AEQ@#$%&^*((^&*~~~````"
        password = "user4"
        roles = ["read-write"]

        try:
            status = self.client.admin_create_user(user, password, roles, policy)
            assert status == 0
            time.sleep(1)
        except Exception:
            pass

        status = self.client.admin_drop_user(user)

        assert status == 0



# Record to select from
client.put(keyTuple, {'bin1': 4, 'bin2': 3})

# Only get bin1
(key, meta, bins) = client.select(keyTuple, ['bin1'])

# Similar output to get()
print(key) # ('test', 'demo', 'key', bytearray(b'...'))
print(meta) # {'ttl': 2592000, 'gen': 1}
print(bins) # {'bin1': 4}

# Get all bins
(key, meta, bins) = client.select(keyTuple, ['bin1', 'bin2'])
print(bins) # {'bin1': 4, 'bin2': 3}

# Get nonexistent bin
(key, meta, bins) = client.select(keyTuple, ['bin3'])
print(bins) # {}



# -*- coding: utf-8 -*-

import pytest

from aerospike import exception as e
from .test_base_class import TestBaseClass
from aerospike_helpers.expressions import (
    And,
    Eq,
    IntBin,
    ListGetByIndex,
    ListGetByIndexRange,
    ListGetByIndexRangeToEnd,
    ListGetByRank,
    ListGetByRankRange,
    ListGetByRankRangeToEnd,
    ListGetByValue,
    ListGetByValueList,
    ListGetByValueRange,
    ListGetByValueRelRankRange,
    ListGetByValueRelRankRangeToEnd,
    ListSize,
    ResultType,
)

import aerospike


@pytest.mark.xfail(TestBaseClass.temporary_xfail(), reason="xfail variable set")
@pytest.mark.usefixtures("as_connection", "connection_config")
class TestGetExpressionBase64(object):
    @pytest.fixture(autouse=True)
    def setup_method(self, request, as_connection):
        pass

    def test_get_expression_base64_pos(self):
        expr = Eq(IntBin("bin1"), 6).compile()
        b64 = self.as_connection.get_expression_base64(expr)
        assert b64 == "kwGTUQKkYmluMQY="

    def test_get_expression_base64_large_expression_pos(self):
        bin = "ilist_bin"
        expr = And(
            Eq(
                ListGetByValueRelRankRange(
                    None,
                    aerospike.LIST_RETURN_COUNT,
                    ListGetByIndex(None, aerospike.LIST_RETURN_VALUE, ResultType.INTEGER, 0, bin),
                    1,
                    3,
                    bin,
                ),
                2,
            ),
            Eq(
                ListGetByValue(
                    None,
                    aerospike.LIST_RETURN_INDEX,
                    6,
                    ListGetByValueRange(None, aerospike.LIST_RETURN_VALUE, 1, 7, bin),
                ),
                [2],
            ),
            Eq(
                ListGetByValueList(
                    None,
                    aerospike.LIST_RETURN_COUNT,
                    [2, 6],
                    ListGetByValueRelRankRangeToEnd(None, aerospike.LIST_RETURN_VALUE, 1, 1, bin),
                ),
                2,
            ),
            Eq(
                ListGetByIndexRangeToEnd(
                    None,
                    aerospike.LIST_RETURN_COUNT,
                    1,
                    ListGetByIndexRange(
                        None,
                        aerospike.LIST_RETURN_VALUE,
                        1,
                        3,
                        bin,
                    ),
                ),
                1,
            ),
            Eq(
                ListGetByRank(
                    None,
                    aerospike.LIST_RETURN_RANK,
                    ResultType.INTEGER,
                    1,
                    ListGetByRankRangeToEnd(None, aerospike.LIST_RETURN_VALUE, 1, bin),
                ),
                1,
            ),
            Eq(ListGetByRankRange(None, aerospike.LIST_RETURN_COUNT, 1, ListSize(None, bin), bin), 2),
        ).compile()

        b64 = self.as_connection.get_expression_base64(expr)
        expected = (
            "lxCTAZV/AgCVGwWVfwIAkxMHAJNRBKlpbGlzdF9iaW4BA5NRBKlpbGlzdF9iaW4CkwGVfwQAkxYBBpV/BAC"
            "UGQcBB5NRBKlpbGlzdF9iaW6SfpECkwGVfwIAkxcFkn6SAgaVfwQAlBsHAQGTUQSpaWxpc3RfYmluApMBlX8CAJMYBQGVfwQAlBgHAQOTU"
            "QSpaWxpc3RfYmluAZMBlX8CAJMVAwGVfwQ"
            "AkxoHAZNRBKlpbGlzdF9iaW4BkwGVfwIAlBoFAZV/AgCREJNRBKlpbGlzdF9iaW6TUQSpaWxpc3RfYmluAg=="
        )

        assert b64 == expected

    def test_get_expression_base64_bad_filter_neg(self):
        with pytest.raises(e.ParamError):
            self.as_connection.get_expression_base64("bad_filter")

    def test_get_expression_base64_bad_list_filter_neg(self):
        with pytest.raises(e.ParamError):
            self.as_connection.get_expression_base64(["hi"])



# -*- coding: utf-8 -*-
import pytest
import random
from aerospike import exception as e


class TestListClear(object):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        keys = []
        for i in range(5):
            key = ("test", "demo", i)
            rec = {"name": "name%s" % (str(i)), "contact_no": [i, i + 1], "city": ["Pune", "Dehli"]}
            as_connection.put(key, rec)
            keys.append(key)

        def teardown():
            """
            Teardown method.
            """
            for key in keys:
                try:
                    as_connection.remove(key)
                except e.RecordNotFound:
                    pass

        request.addfinalizer(teardown)

    def test_pos_list_clear_with_correct_paramters(self):
        """
        Invoke list_clear() with correct parameters
        """
        key = ("test", "demo", 1)

        status = self.as_connection.list_clear(key, "contact_no")

        assert status == 0

        (key, _, bins) = self.as_connection.get(key)
        assert bins == {"city": ["Pune", "Dehli"], "contact_no": [], "name": "name1"}

    def test_pos_list_clear_list_with_correct_policy(self):
        """
        Invoke list_clear() removes all list elements with correct policy
        """

        key = ("test", "demo", 2)

        status = self.as_connection.list_clear(key, "city")
        assert status == 0

        (key, _, bins) = self.as_connection.get(key)
        assert bins == {"city": [], "contact_no": [2, 3], "name": "name2"}

    # Negative Tests
    def test_neg_list_clear_with_incorrect_policy(self):
        """
        Invoke list_clear() with incorrect policy
        """
        key = ("test", "demo", 1)
        policy = {"timeout": 0.5}
        try:
            self.as_connection.list_clear(key, "contact_no", {}, policy)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "timeout is invalid"

    def test_neg_list_clear_with_nonexistent_key(self):
        """
        Invoke list_clear() with non-existent key
        """
        if self.server_version < [3, 15, 2]:
            pytest.skip("Change of error beginning in 3.15")
        key = ("test", "demo", "test_neg_list_clear_with_nonexistent_key")
        with pytest.raises(e.RecordNotFound):
            self.as_connection.list_clear(key, "contact_no")

    def test_neg_list_clear_with_nonexistent_bin(self):
        """
        Invoke list_clear() with non-existent bin
        """
        key = ("test", "demo", 1)
        charSet = "abcdefghijklmnopqrstuvwxyz1234567890"
        minLength = 5
        maxLength = 10
        length = random.randint(minLength, maxLength)
        bin = "".join(map(lambda unused: random.choice(charSet), range(length))) + ".com"
        try:
            self.as_connection.list_clear(key, bin)
        except e.BinIncompatibleType as exception:
            assert exception.code == 12

    def test_neg_list_clear_with_extra_parameter(self):
        """
        Invoke list_clear() with extra parameter.
        """
        key = ("test", "demo", 1)
        policy = {"timeout": 1000}
        with pytest.raises(TypeError) as typeError:
            self.as_connection.list_clear(key, "contact_no", {}, policy, "")

        assert "list_clear() takes at most 4 arguments (5 given)" in str(typeError.value)

    def test_neg_list_clear_policy_is_string(self):
        """
        Invoke list_clear() with policy is string
        """
        key = ("test", "demo", 1)
        try:
            self.as_connection.list_clear(key, "contact_no", {}, "")

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "policy must be a dict"

    def test_neg_list_clear_key_is_none(self):
        """
        Invoke list_clear() with key is none
        """
        try:
            self.as_connection.list_clear(None, "contact_no")

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "key is invalid"

    def test_neg_list_clear_bin_is_none(self):
        """
        Invoke list_clear() with bin is none
        """
        key = ("test", "demo", 1)
        try:
            self.as_connection.list_clear(key, None)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Bin name should be of type string"

    def test_neg_list_clear_meta_type_integer(self):
        """
        Invoke list_clear() with metadata input is of type integer
        """
        key = ("test", "demo", 1)
        try:
            self.as_connection.list_clear(key, "contact_no", 888)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Metadata should be of type dictionary"

    def test_neg_list_clear_with_no_parameters(self):
        """
        Invoke list_clear() without any mandatory parameters.
        """
        with pytest.raises(TypeError) as typeError:
            self.as_connection.list_clear()
        assert "argument 'key' (pos 1)" in str(typeError.value)



# These fail if we don't have a new server
# -*- coding: utf-8 -*-
import pytest
import aerospike
from aerospike import exception as e
from aerospike_helpers.operations import list_operations as lo
from aerospike_helpers.operations import map_operations as map_ops


def get_list_result_from_operation(client, key, operation, binname):
    """
    Just perform a single operation and return the bins portion of the
    result
    """
    _, _, result_bins = client.operate(key, [operation])
    return result_bins[binname]


@pytest.mark.xfail(reason="This requires server 4.3.13")
class TestNewRelativeCDTValues(object):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        """
        Setup Method
        """
        self.keys = []
        cdt_list_val = [[0, "a"], [1, "b"], [1, "c"], [1, "d", "e"], [2, "f"], [2, "two"], [3, "g"]]
        cdt_map_val = {
            "a": [0, "a"],
            "b": [1, "b"],
            "c": [1, "c"],
            "d": [1, "d", "e"],
            "e": [2, "f"],
            "f": [2, "two"],
            "g": [3, "g"],
        }

        self.cdt_key = ("test", "cdt_values", "wildcard")
        self.cdt_list_bin = "cdt_list_bin"
        self.cdt_map_bin = "cdt_map_bin"

        self.as_connection.put(self.cdt_key, {self.cdt_list_bin: cdt_list_val, self.cdt_map_bin: cdt_map_val})
        # Make sure the list is ordered, in order to get expected return order.
        ops = [lo.list_sort(self.cdt_list_bin, 0), lo.list_set_order(self.cdt_list_bin, aerospike.LIST_ORDERED)]
        self.as_connection.operate(self.cdt_key, ops)

        self.keys.append(self.cdt_key)

        yield

        for rec_key in self.keys:
            try:
                self.as_connection.remove(rec_key)
            except e.AerospikeError:
                pass

    def test_cdt_wild_card_list_value_multi_element(self):
        operation = lo.list_get_by_value(self.cdt_list_bin, [1, aerospike.CDTWildcard()], aerospike.LIST_RETURN_VALUE)

        result = get_list_result_from_operation(self.as_connection, self.keys[0], operation, self.cdt_list_bin)

        # All items starting with 1
        assert len(result) == 3
        for lst in result:
            assert lst[0] == 1

    def test_cdt_wild_card_list_value(self):
        # This is does gthe value match [*]
        operation = lo.list_get_by_value(self.cdt_list_bin, aerospike.CDTWildcard(), aerospike.LIST_RETURN_VALUE)

        result = get_list_result_from_operation(self.as_connection, self.keys[0], operation, self.cdt_list_bin)

        # All items starting with 1
        assert len(result) == 7

    def test_cdt_infinite_list_range_value(self):
        operation = lo.list_get_by_value_range(
            self.cdt_list_bin, aerospike.LIST_RETURN_VALUE, [1, aerospike.null()], [1, aerospike.CDTInfinite()]
        )

        result = get_list_result_from_operation(self.as_connection, self.keys[0], operation, self.cdt_list_bin)

        # All items starting with 1
        assert len(result) == 3
        for lst in result:
            assert lst[0] == 1

    def test_map_value_wildcard(self):
        operation = map_ops.map_get_by_value(self.cdt_map_bin, aerospike.CDTWildcard(), aerospike.MAP_RETURN_KEY)
        result = get_list_result_from_operation(self.as_connection, self.cdt_key, operation, self.cdt_map_bin)

        assert set(result) == set(("a", "b", "c", "d", "e", "f", "g"))

    # TODO: duplicate test names
    def test_map_value_wildcard1(self):
        operation = map_ops.map_get_by_value(
            self.cdt_map_bin,
            [1, aerospike.CDTWildcard()],
            aerospike.MAP_RETURN_KEY,
        )
        result = get_list_result_from_operation(self.as_connection, self.cdt_key, operation, self.cdt_map_bin)

        assert set(result) == set(("b", "c", "d"))

    def test_map_value_null_infinity_range(self):
        operation = map_ops.map_get_by_value_range(
            self.cdt_map_bin, [1, aerospike.null()], [1, aerospike.CDTInfinite()], aerospike.MAP_RETURN_KEY
        )
        result = get_list_result_from_operation(self.as_connection, self.cdt_key, operation, self.cdt_map_bin)

        assert set(result) == set(("b", "c", "d"))



# -*- coding: utf-8 -*-

import pytest

from .test_base_class import TestBaseClass
from .as_status_codes import AerospikeStatus
import aerospike
from aerospike_helpers import expressions as exp
from aerospike import exception as e


"""
Summary of functions:
Sample:
list_append(binname, val): appends an item to the end of a list
list_append_extra(binname, val, extra_arg): same as above, extra arg is unused

TestRecord:
bin_udf_operation_integer(record, bin_name, x, y):
    adds x + y to record[bin_name]
bin_udf_operation_string(record, bin_name, str):
    concatenates str to record[bin_name]
bin_udf_operation_bool(record, bin_name): returns record[bin_name]
list_iterate(record, bin, index_of_ele): equivalent to:
    record[bin].append(58)
    record[bin][index_of_ele] = 222
list_iterate_returns_list(record, bin, index_of_ele):
    Same as previous but returns the modified list

map_iterate(record, bin, set_value): Changes every value
    in the map map[bin] to be equal to set_value

map_iterate_returns_map(record, bin, set_value):
    Same as above but returns the map
function udf_returns_record(rec):
    Return a map representation of a record
udf_without_arg_and_return(record):
    does nothing and returns nothing
function udf_put_bytes(record, bin):
    creates an array of bytes and inserts it in record[bin]

BasicOps:

create_record(rec, value)
    creates a new record with the data
        record['bin'] = value
"""


def add_indexes_and_udfs(client):
    """
    Load the UDFs used in the tests and setup indexes
    """
    policy = {}
    try:
        client.index_integer_create("test", "demo", "age", "age_index", policy)
    except e.IndexFoundError:
        pass
    try:
        client.index_integer_create("test", "demo", "age1", "age_index1", policy)
    except e.IndexFoundError:
        pass

    udf_type = 0
    udf_files = ("sample.lua", "test_record_udf.lua", "udf_basic_ops.lua")

    for module in udf_files:
        client.udf_put(module, udf_type, policy)


def remove_indexes_and_udfs(client):
    """
    Remove all of the UDFS and indexes created for these tests
    """
    policy = {}

    try:
        client.index_remove("test", "age_index", policy)
    except e.IndexNotFound:
        pass

    try:
        client.index_remove("test", "age_index1", policy)
    except e.IndexNotFound:
        pass

    udf_files = ("sample.lua", "test_record_udf.lua", "udf_basic_ops.lua")

    for module in udf_files:
        client.udf_remove(module, policy)


class TestApply(TestBaseClass):
    def setup_class(cls):
        # Register setup and teardown functions
        cls.connection_setup_functions = [add_indexes_and_udfs]
        cls.connection_teardown_functions = [remove_indexes_and_udfs]

    @pytest.fixture(autouse=True)
    def setup(self, request, connection_with_config_funcs):
        as_connection = connection_with_config_funcs

        for i in range(5):
            key = ("test", "demo", i)
            rec = {
                "name": ["name%s" % (str(i))],
                "addr": "name%s" % (str(i)),
                "age": i,
                "no": i,
                "basic_map": {"k30": 6, "k20": 5, "k10": 1},
            }
            as_connection.put(key, rec)

        def teardown():
            for i in range(5):
                key = ("test", "demo", i)
                as_connection.remove(key)

        request.addfinalizer(teardown)

    @pytest.mark.parametrize(
        "func_args, test_bin, expected",
        (
            (["name", "car"], "name", ["name1", "car"]),
            (["name", None], "name", ["name1", None]),
            (["name", 1], "name", ["name1", 1]),
            (["name", [1, 2]], "name", ["name1", [1, 2]]),
            (["name", {"a": "b"}], "name", ["name1", {"a": "b"}]),
        ),
        ids=["string", "None", "Integer", "list", "dict"],
    )
    def test_apply_causing_list_append_with_correct_params(self, func_args, test_bin, expected):

        key = ("test", "demo", 1)
        retval = self.as_connection.apply(key, "sample", "list_append", func_args)

        _, _, bins = self.as_connection.get(key)

        assert bins[test_bin] == expected
        assert retval == 0  # the list_append UDF returns 0

    @pytest.mark.parametrize(
        "func_args, test_bin, expressions, expected",
        ((["name", 1], "name", exp.Eq(exp.IntBin("age"), 1), ["name1", 1]),),
        ids=[
            "Integer",
        ],
    )
    def test_apply_causing_list_append_with_correct_params_with_expressions(
        self, func_args, test_bin, expressions, expected
    ):

        key = ("test", "demo", 1)
        retval = self.as_connection.apply(
            key, "sample", "list_append", func_args, policy={"expressions": expressions.compile()}
        )

        _, _, bins = self.as_connection.get(key)

        assert bins[test_bin] == expected
        assert retval == 0  # the list_append UDF returns 0

    def test_apply_return_bool_true(self):
        """
        Invoke apply() with UDF that will return booleans.
        """
        key = ("test", "demo", 1)
        retval = self.as_connection.apply(key, "test_record_udf", "bool_check", [])
        assert retval is True

    def test_apply_return_bool_false(self):
        """
        Invoke apply() with UDF that will return booleans.
        """
        key = ("test", "demo", "non_existent_record")
        retval = self.as_connection.apply(key, "test_record_udf", "bool_check", [])
        assert retval is False

    def test_apply_operations_on_map(self):
        """
        Invoke apply() with map
        """
        key = ("test", "demo", 1)
        retval = self.as_connection.apply(key, "test_record_udf", "map_iterate", ["basic_map", 555])
        assert retval is None
        # Since this udf changes all of the record values to be 555,
        # all of the values in the 'basic_map' should be 555
        (key, _, bins) = self.as_connection.get(key)
        assert bins["basic_map"] == {"k30": 555, "k20": 555, "k10": 555}

    def test_apply_with_correct_parameters_float_argument(self):
        """
        Invoke apply() with correct arguments with a floating value in the
        list of arguments
        """
        if self.skip_old_server is True:
            pytest.skip("Server does not support apply on float type as lua argument")
        key = ("test", "demo", 1)
        retval = self.as_connection.apply(key, "sample", "list_append", ["name", 5.434])
        _, _, bins = self.as_connection.get(key)

        assert bins["name"] == ["name1", 5.434]
        assert retval == 0

    def test_apply_with_policy(self):
        """
        Invoke apply() with policy
        """
        policy = {"total_timeout": 1000}
        key = ("test", "demo", 1)
        retval = self.as_connection.apply(key, "sample", "list_append", ["name", "car"], policy)
        (key, _, bins) = self.as_connection.get(key)
        assert retval == 0
        assert bins["name"] == ["name1", "car"]

    def test_apply_with_integer(self):
        """
        Invoke apply() with integer
        """
        key = ("test", "demo", 1)
        retval = self.as_connection.apply(key, "test_record_udf", "bin_udf_operation_integer", ["age", 2, 20])

        assert retval == 23
        (key, _, bins) = self.as_connection.get(key)
        assert bins["age"] == 23

    def test_apply_with_operation_on_string(self):
        """
        Invoke apply() with string
        """
        key = ("test", "demo", 1)
        retval = self.as_connection.apply(key, "test_record_udf", "bin_udf_operation_string", ["addr", " world"])

        assert retval == "name1 world"
        (key, _, bins) = self.as_connection.get(key)
        assert bins["addr"] == "name1 world"

    def test_apply_with_function_returning_record(self):
        """
        Invoke apply() with record
        """
        key = ("test", "demo", 1)
        retval = self.as_connection.apply(key, "test_record_udf", "udf_returns_record", [])

        # This function returns a record which gets mapped to a python dict
        assert retval is not None
        assert isinstance(retval, dict)

    def test_apply_using_bytearray(self):
        """
        Invoke apply() with a bytearray as a argument
        """
        key = ("test", "demo", "apply_insert")
        self.as_connection.apply(key, "udf_basic_ops", "create_record", [bytearray("asd;as[d'as;d", "utf-8")])
        (key, _, bins) = self.as_connection.get(key)

        assert bins == {"bin": bytearray(b"asd;as[d'as;d")}

        self.as_connection.remove(key)

    def test_apply_with_extra_argument_to_lua_function(self):
        """
        Invoke apply() with extra argument to lua
        Should not raise an error
        """
        key = ("test", "demo", 1)
        retval = self.as_connection.apply(key, "sample", "list_append", ["name", "car", 1])
        assert retval == 0
        (key, _, bins) = self.as_connection.get(key)
        assert bins["name"] == ["name1", "car"]

    def test_apply_with_missing_argument_to_lua(self):
        """
        Invoke apply() passing less than the ex arguments
        to a lua function. Should not raise an error
        """
        key = ("test", "demo", 1)
        retval = self.as_connection.apply(key, "sample", "list_append_extra", ["name", "car"])
        assert retval == 0
        (key, _, bins) = self.as_connection.get(key)
        assert bins["name"] == ["name1", "car"]

    def test_apply_with_unicode_module_and_function(self):
        """
        Invoke apply() with unicode module and function
        """
        key = ("test", "demo", 1)
        retval = self.as_connection.apply(key, "sample", "list_append", ["name", "car"])
        (key, _, bins) = self.as_connection.get(key)

        assert bins["name"] == ["name1", "car"]
        assert retval == 0

    def test_apply_with_correct_parameters_without_connection(self):
        """
        Invoke apply() with correct arguments without connection
        """
        key = ("test", "demo", 1)
        config = TestBaseClass.get_connection_config()
        client1 = aerospike.client(config)
        client1.close()

        with pytest.raises(e.ClusterError) as err_info:
            client1.apply(key, "sample", "list_append", ["name", "car"])

        err_code = err_info.value.code

        assert err_code == AerospikeStatus.AEROSPIKE_CLUSTER_ERROR

    def test_apply_with_arg_causing_error(self):
        """
        Invoke apply() with ia string argument when integer is required
        """
        key = ("test", "demo", 1)

        with pytest.raises(e.UDFError) as err_info:
            self.as_connection.apply(key, "test_record_udf", "bin_udf_operation_integer", ["age", "not an", "integer"])

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_UDF

    def test_apply_with_no_parameters(self):
        """
        Invoke apply() without any mandatory parameters.
        This should raise an error
        """
        with pytest.raises(TypeError) as typeError:
            self.as_connection.apply()

        assert "argument 'key' (pos 1)" in str(typeError.value)

    def test_apply_with_no_argument_in_lua(self):
        """
        Call apply without the args parameter
        """
        key = ("test", "demo", 1)
        with pytest.raises(TypeError) as typeError:
            self.as_connection.apply(key, "sample", "list_append_extra")
        assert "argument 'args' (pos 4)" in str(typeError.value)

    def test_apply_with_incorrect_policy(self):
        """
        Invoke apply() with incorrect policy
        """
        policy = {"total_timeout": 0.1}
        key = ("test", "demo", 1)
        with pytest.raises(e.ParamError) as err_info:
            self.as_connection.apply(key, "sample", "list_append", ["name", "car"], policy)

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_apply_with_extra_argument(self):
        """
        Invoke apply() with extra argument.
        """
        policy = {"total_timeout": 1000}
        key = ("test", "demo", 1)
        with pytest.raises(TypeError) as typeError:
            self.as_connection.apply(key, "sample", "list_append", ["name", "car"], policy, "")

        assert "apply() takes at most 5 arguments (6 given)" in str(typeError.value)

    @pytest.mark.parametrize(
        "module, function, arguments",
        (
            ("sample", "list_append", ["addr", "car"]),
            ("", "", ["name", "car"]),
            ("wrong_module_name", "list_append", ["name", "car"]),
            ("sample", "list_prepend", ["name", "car"]),
        ),
        ids=["incorrect argument type", "empty module and function", "non existent module", "non existent function"],
    )
    def test_udf_error_causing_args(self, module, function, arguments):
        key = ("test", "demo", 1)

        with pytest.raises(e.UDFError) as err_info:
            self.as_connection.apply(key, module, function, arguments)

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_UDF

    def test_apply_with_key_as_string(self):
        """
        Invoke apply() with key as string
        """
        with pytest.raises(e.ParamError) as err_info:
            self.as_connection.apply("", "sample", "list_append", ["name", "car"])

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_apply_with_key_as_none(self):
        """
        Invoke apply() with key as none
        """
        with pytest.raises(e.ParamError) as err_info:
            self.as_connection.apply(None, "sample", "list_append", ["name", "car"])

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_apply_with_incorrect_ns_set(self):
        """
        Invoke apply() with incorrect ns and set
        """

        with pytest.raises(e.ClientError):
            key = ("test1", "demo", 1)
            self.as_connection.apply(key, "sample", "list_prepend", ["name", "car"])



import aerospike
from aerospike_helpers.operations import map_operations as mop
from aerospike_helpers.operations import list_operations as lop
import aerospike_helpers.cdt_ctx as ctx
from aerospike import KeyOrderedDict

config = { 'hosts': [ ("localhost", 3000), ] }
client = aerospike.client(config).connect()
map_policy={'map_order': aerospike.MAP_KEY_VALUE_ORDERED}

key = ("test", "demo", 100)
client.put(key, {'map_list': []})

map_ctx1 = ctx.cdt_ctx_list_index(0)
map_ctx2 = ctx.cdt_ctx_list_index(1)
map_ctx3 = ctx.cdt_ctx_list_index(2)

my_dict1 = {'a': 1, 'b': 2, 'c': 3}
my_dict2 = {'d': 4, 'e': 5, 'f': 6}
my_dict3 = {'g': 7, 'h': 8, 'i': 9}

ops = [ 
        lop.list_append_items('map_list', [my_dict1, my_dict2, my_dict3]),
        mop.map_set_policy('map_list', map_policy, [map_ctx1]),
        mop.map_set_policy('map_list', map_policy, [map_ctx2]),
        mop.map_set_policy('map_list', map_policy, [map_ctx3])
    ]
client.operate(key, ops)

_, _, res = client.get(key)
print(res)

element = KeyOrderedDict({'f': 6, 'e': 5, 'd': 4}) # this will match my_dict2 because it will be converted to key ordered.

ops = [ 
        lop.list_get_by_value('map_list', element, aerospike.LIST_RETURN_COUNT)
    ]
_, _, res = client.operate(key, ops)
print(res)

client.remove(key)
client.close()

# EXPECTED OUTPUT:
# {'map_list': [{'a': 1, 'b': 2, 'c': 3}, {'d': 4, 'e': 5, 'f': 6}, {'g': 7, 'h': 8, 'i': 9}]}
# {'map_list': 1}



# -*- coding: utf-8 -*-


import pytest

import aerospike

from aerospike_helpers import expressions as exp
from aerospike_helpers.batch import records as br
from aerospike_helpers.operations import operations as op
from aerospike_helpers.operations import list_operations as lop
from aerospike import exception as e
from .test_base_class import TestBaseClass
from .as_status_codes import AerospikeStatus


def add_udfs(client):
    """
    Load the UDFs used in the tests
    """
    policy = {}
    udf_type = 0
    udf_files = ("sample.lua", "test_record_udf.lua", "udf_basic_ops.lua")

    for module in udf_files:
        client.udf_put(module, udf_type, policy)


def remove_udfs(client):
    """
    Remove all of the UDFS created for these tests
    """
    policy = {}

    udf_files = ("sample.lua", "test_record_udf.lua", "udf_basic_ops.lua")

    for module in udf_files:
        client.udf_remove(module, policy)


class TestBatchWrite(TestBaseClass):
    def setup_class(cls):
        # Register setup and teardown functions
        cls.connection_setup_functions = [add_udfs]
        cls.connection_teardown_functions = [remove_udfs]

    @pytest.fixture(autouse=True)
    def setup(self, request, connection_with_config_funcs):
        as_connection = connection_with_config_funcs

        if self.server_version < [6, 0]:
            pytest.mark.xfail(reason="Servers older than 6.0 do not support batch writes.")
            pytest.xfail()

        self.test_ns = "test"
        self.test_set = "demo"
        self.keys = []
        self.batch_size = 5

        for i in range(self.batch_size):
            key = ("test", "demo", i)
            rec = {
                "count": i,
                "name": "name10",
                "t": True,
                "age": 10,
                "balance": 100,
                "key": 10,
                "ilist_bin": [
                    1,
                    2,
                    6,
                ],
                "imap_bin": {
                    1: 1,
                    2: 2,
                    3: 6,
                },
            }
            as_connection.put(key, rec)
            self.keys.append(key)

        def teardown():
            for i in range(self.batch_size):
                key = ("test", "demo", i)
                try:
                    as_connection.remove(key)
                except e.RecordNotFound:
                    # records may have been removed by out of order remove batch ops
                    pass

        request.addfinalizer(teardown)

    @pytest.mark.parametrize(
        "name, batch_records, policy, exp_res, exp_rec",
        [
            (
                "simple-write",
                br.BatchRecords([br.Write(("test", "demo", 1), [op.write("new", 10), op.read("new")])]),
                {},
                [AerospikeStatus.AEROSPIKE_OK],
                [{"new": 10}],
            ),
            (
                "write-with-batch-policy",
                br.BatchRecords(
                    [
                        br.Write(
                            ("test", "demo", 1),
                            [op.write("new", 10), op.read("new")],
                        )
                    ]
                ),
                {"total_timeout": 2000, "max_retries": 2, "allow_inline_ssd": True, "respond_all_keys": False},
                [AerospikeStatus.AEROSPIKE_OK],
                [{"new": 10}],
            ),
            (
                "write-with-policy",
                br.BatchRecords(
                    [
                        br.Write(
                            ("test", "demo", 1),
                            [op.write("new", 10), op.read("new")],
                            meta={"gen": 1, "ttl": aerospike.TTL_NEVER_EXPIRE},
                            policy={
                                "key": aerospike.POLICY_KEY_SEND,
                                "commit_level": aerospike.POLICY_COMMIT_LEVEL_MASTER,
                                "gen": aerospike.POLICY_GEN_IGNORE,
                                "exists": aerospike.POLICY_EXISTS_UPDATE,
                                "durable_delete": False,
                                "expressions": exp.Eq(exp.IntBin("count"), 1).compile(),
                            },
                        )
                    ]
                ),
                {},
                [AerospikeStatus.AEROSPIKE_OK],
                [{"new": 10}],
            ),
            (
                "simple-read",
                br.BatchRecords(
                    [
                        br.Read(
                            ("test", "demo", 1),
                            [op.read("count")],
                        )
                    ]
                ),
                {},
                [AerospikeStatus.AEROSPIKE_OK],
                [{"count": 1}],
            ),
            (
                "read-all-bins",
                br.BatchRecords([br.Read(("test", "demo", 1), ops=None, read_all_bins=True)]),
                {},
                [AerospikeStatus.AEROSPIKE_OK],
                [
                    {
                        "count": 1,
                        "name": "name10",
                        "t": True,
                        "age": 10,
                        "balance": 100,
                        "key": 10,
                        "ilist_bin": [
                            1,
                            2,
                            6,
                        ],
                        "imap_bin": {
                            1: 1,
                            2: 2,
                            3: 6,
                        },
                    }
                ],
            ),
            (
                "read-with-policy",
                br.BatchRecords(
                    [
                        br.Write(
                            ("test", "demo", 1),
                            [op.write("new", 10), op.read("new")],
                            meta={"gen": 1, "ttl": aerospike.TTL_NEVER_EXPIRE},
                            policy={
                                "read_mode_ap": aerospike.POLICY_READ_MODE_AP_ONE,
                                "expressions": exp.Eq(exp.IntBin("count"), 1).compile(),
                            },
                        )
                    ]
                ),
                {},
                [AerospikeStatus.AEROSPIKE_OK],
                [{"new": 10}],
            ),
            (
                "simple-remove",
                br.BatchRecords(
                    [
                        br.Remove(("test", "demo", 1)),
                        br.Write(
                            ("test", "demo", 1),
                            [op.write("new", 10), op.read("new")],
                        ),
                    ]
                ),
                {},
                [AerospikeStatus.AEROSPIKE_OK, AerospikeStatus.AEROSPIKE_OK],
                [{}, {"new": 10}],
            ),
            (
                "remove-with-policy",
                br.BatchRecords(
                    [
                        br.Remove(
                            ("test", "demo", 1),
                            policy={
                                "key": aerospike.POLICY_KEY_SEND,
                                "commit_level": aerospike.POLICY_COMMIT_LEVEL_MASTER,
                                "gen": aerospike.POLICY_GEN_IGNORE,
                                "durable_delete": False,
                                "expressions": exp.Eq(exp.IntBin("count"), 1).compile(),
                            },
                        ),
                        br.Write(("test", "demo", 1), [op.write("new", 10), op.read("new")]),
                    ]
                ),
                {},
                [AerospikeStatus.AEROSPIKE_OK, AerospikeStatus.AEROSPIKE_OK],
                [{}, {"new": 10}],
            ),
            (
                "simple-apply",
                br.BatchRecords(
                    [
                        br.Apply(("test", "demo", 1), "sample", "list_append", ["ilist_bin", 200]),
                        br.Read(
                            ("test", "demo", 1),
                            [
                                lop.list_get_by_rank("ilist_bin", 0, aerospike.LIST_RETURN_VALUE),
                            ],
                        ),
                    ]
                ),
                {},
                [AerospikeStatus.AEROSPIKE_OK, AerospikeStatus.AEROSPIKE_OK],
                [{"SUCCESS": 0}, {"ilist_bin": 1}],
            ),
            (
                "apply-with-policy",
                br.BatchRecords(
                    [
                        br.Apply(
                            ("test", "demo", 1),
                            "sample",
                            "list_append",
                            ["ilist_bin", 200],
                            policy={
                                "key": aerospike.POLICY_KEY_DIGEST,
                                "commit_level": aerospike.POLICY_COMMIT_LEVEL_MASTER,
                                "ttl": aerospike.TTL_NEVER_EXPIRE,
                                "durable_delete": False,
                                "expressions": exp.Eq(exp.IntBin("count"), 1).compile(),
                            },
                        ),
                        br.Read(
                            ("test", "demo", 1),
                            [
                                lop.list_get_by_rank("ilist_bin", 0, aerospike.LIST_RETURN_VALUE),
                            ],
                        ),
                    ]
                ),
                {},
                [AerospikeStatus.AEROSPIKE_OK, AerospikeStatus.AEROSPIKE_OK],
                [{"SUCCESS": 0}, {"ilist_bin": 1}],
            ),
            (
                "write-read",
                br.BatchRecords(
                    [
                        br.Write(("test", "demo", 1), [op.write("new", 11), op.read("new")]),
                        br.Read(
                            ("test", "demo", 1),
                            [lop.list_get_by_rank("ilist_bin", -1, aerospike.LIST_RETURN_VALUE), op.read("balance")],
                        ),
                    ]
                ),
                {},
                [AerospikeStatus.AEROSPIKE_OK, AerospikeStatus.AEROSPIKE_OK],
                [{"new": 11}, {"ilist_bin": 6, "balance": 100}],
            ),
            (
                "complex",
                br.BatchRecords(
                    [
                        br.Write(
                            ("test", "demo", 1),
                            [
                                op.write("ilist_bin", [2, 6]),
                                op.write("balance", 100),
                                op.read("ilist_bin"),
                                op.read("balance"),
                            ],
                        ),
                        br.Read(
                            ("test", "demo", 1), [lop.list_get_by_rank("ilist_bin", -1, aerospike.LIST_RETURN_VALUE)]
                        ),
                        br.Apply(
                            ("test", "demo", 3),
                            "sample",
                            "list_append",
                            ["ilist_bin", 200],
                            policy={"expressions": exp.Eq(exp.IntBin("count"), 3).compile()},
                        ),
                        br.Read(
                            ("test", "demo", 3),
                            [lop.list_get_by_rank("ilist_bin", 0, aerospike.LIST_RETURN_VALUE), op.read("balance")],
                        ),
                    ]
                ),
                {},
                [
                    AerospikeStatus.AEROSPIKE_OK,
                    AerospikeStatus.AEROSPIKE_OK,
                    AerospikeStatus.AEROSPIKE_OK,
                    AerospikeStatus.AEROSPIKE_OK,
                    AerospikeStatus.AEROSPIKE_OK,
                ],
                [
                    {"balance": 100, "ilist_bin": [2, 6]},
                    {"ilist_bin": 6},
                    {"SUCCESS": 0},
                    {"balance": 100, "ilist_bin": 1},
                ],
            ),
            (
                "read-many",
                br.BatchRecords(
                    [
                        br.Read(
                            ("test", "demo", i),
                            [
                                op.read("count"),
                            ],
                        )
                        for i in range(5)
                    ]
                ),
                {},
                [
                    AerospikeStatus.AEROSPIKE_OK,
                    AerospikeStatus.AEROSPIKE_OK,
                    AerospikeStatus.AEROSPIKE_OK,
                    AerospikeStatus.AEROSPIKE_OK,
                    AerospikeStatus.AEROSPIKE_OK,
                ],
                [
                    {"count": 0},
                    {"count": 1},
                    {"count": 2},
                    {"count": 3},
                    {"count": 4},
                ],
            ),
        ],
    )
    def test_batch_write_pos(self, name, batch_records, policy, exp_res, exp_rec):
        """
        Test batch_write positive
        """
        res = self.as_connection.batch_write(batch_records, policy)

        for i, batch_rec in enumerate(res.batch_records):
            # print("name:", name)
            assert batch_rec.result == exp_res[i]
            assert batch_rec.record[2] == exp_rec[i]

    @pytest.mark.parametrize(
        "name, batch_records, policy, exp_res",
        [
            ("bad-batch-records", ["bad", "batch", "records"], {}, e.ParamError),
            (
                "bad-batch-record",
                br.BatchRecords(
                    [
                        br.Read(
                            ("test", "demo", 1),
                            [
                                op.read("count"),
                            ],
                        ),
                        "bad_batch_record",
                    ]
                ),
                {},
                e.ParamError,
            ),
            (
                "bad-batch-record-key",
                br.BatchRecords(
                    [
                        br.Read(
                            "bad_key",
                            [
                                op.read("count"),
                            ],
                        )
                    ]
                ),
                {},
                e.ParamError,
            ),
            ("bad-batch-record-ops", br.BatchRecords([br.Read(("test", "demo", 1), {"bad": "ops"})]), {}, e.ParamError),
            (
                "bad-batch-record-policy",
                br.BatchRecords(
                    [
                        br.Read(
                            ("test", "demo", 1),
                            [
                                op.read("count"),
                            ],
                            meta=None,
                            policy="bad policy",
                        )
                    ]
                ),
                {},
                e.ParamError,
            ),
            (
                "bad-batch-policy",
                br.BatchRecords(
                    [
                        br.Read(
                            ("test", "demo", 1),
                            [
                                op.read("count"),
                            ],
                        )
                    ]
                ),
                "bad policy",
                e.ParamError,
            ),
        ],
    )
    def test_batch_write_neg(self, name, batch_records, policy, exp_res):
        """
        Test batch_write positive
        """

        with pytest.raises(exp_res):
            self.as_connection.batch_write(batch_records, policy)

    def test_batch_write_neg_connection(self):
        """
        Test batch_write negative with bad connection.
        """

        batch_records = []

        exp_res = e.ClientError

        with pytest.raises(exp_res):
            bad_client = aerospike.client({"hosts": [("bad_addr", 3000)]})
            bad_client.batch_write(batch_records)



# -*- coding: utf-8 -*-
import pytest
import random
from datetime import datetime
from aerospike import exception as e
from aerospike_helpers.operations import bitwise_operations

import aerospike

random.seed(datetime.now())


class TestBitwiseOperations(object):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        """
        Setup Method
        """
        self.keys = []
        self.five_zero_blob = bytearray([0] * 5)
        self.five_one_blob = bytearray([1] * 5)
        self.zero_one_blob = bytearray([0] * 5 + [1] * 5)
        self.count_blob = bytearray([1] * 1 + [86] * 2 + [255] * 1 + [3] * 1)
        self.five_255_blob = bytearray([255] * 5)

        self.test_key = "test", "demo", "bitwise_op"
        self.test_bin_zeroes = "bitwise0"
        self.test_bin_ones = "bitwise1"
        self.zero_one_bin = "bitwise01"
        self.count_bin = "count"
        self.five_255_bin = "255"
        self.as_connection.put(
            self.test_key,
            {
                self.test_bin_zeroes: self.five_zero_blob,
                self.test_bin_ones: self.five_one_blob,
                self.zero_one_bin: self.zero_one_blob,
                self.count_bin: self.count_blob,
                self.five_255_bin: self.five_255_blob,
            },
        )
        self.keys.append(self.test_key)

        yield

        for key in self.keys:
            try:
                self.as_connection.remove(key)
            except e.AerospikeError:
                pass

    def test_bit_resize_defaults(self):
        """
        Perform a bit_resize operation with default flags.
        """
        ops = [bitwise_operations.bit_resize(self.test_bin_ones, 10)]
        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        #  We should not have changed the zeroes bin
        assert bins[self.test_bin_zeroes] == self.five_zero_blob

        assert len(bins[self.test_bin_ones]) == 10
        # We expect the newly added zeroes to be added to the end of the bytearray
        assert bins[self.test_bin_ones] == bytearray([1] * 5 + [0] * 5)

    def test_bit_resize_from_front(self):
        """
        Perform a bit_resize operation with resize from front flags.
        """
        ops = [bitwise_operations.bit_resize(self.test_bin_ones, 10, resize_flags=aerospike.BIT_RESIZE_FROM_FRONT)]
        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        #  We should not have changed the zeroes bin
        assert bins[self.test_bin_zeroes] == self.five_zero_blob

        assert len(bins[self.test_bin_ones]) == 10
        # We expect the newly added zeroes to be added to the front of the bytearray
        assert bins[self.test_bin_ones] == bytearray([0] * 5 + [1] * 5)

    def test_bit_resize_grow_only_allows_grow(self):
        """
        Perform a bit_resize operation with grow only resize flags.
        """
        ops = [bitwise_operations.bit_resize(self.test_bin_ones, 10, resize_flags=aerospike.BIT_RESIZE_GROW_ONLY)]
        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)

        assert len(bins[self.test_bin_ones]) == 10

    def test_bit_resize_shrink_only_allows_shrink(self):
        """
        Perform a bit_resize operation with shrink only resize flags.
        """
        ops = [bitwise_operations.bit_resize(self.test_bin_ones, 1, resize_flags=aerospike.BIT_RESIZE_SHRINK_ONLY)]
        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)

        assert len(bins[self.test_bin_ones]) == 1

    def test_bit_resize_shrink_removes_from_end(self):
        """
        Shrinking a bytearray [0, 0, 0, 0, 0, 1, 1, 1, 1, 1]
        should leave:
        [0, 0, 0, 0, 0]
        """
        ops = [bitwise_operations.bit_resize(self.zero_one_bin, 5)]
        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)

        assert len(bins[self.zero_one_bin]) == 5
        assert bins[self.zero_one_bin] == bytearray([0] * 5)

    def test_bit_resize_shrink_from_front(self):
        """
        Shrinking a bytearray [0, 0, 0, 0, 0, 1, 1, 1, 1, 1]
        from the front should leave:
        [1, 1, 1, 1, 1]
        """
        ops = [bitwise_operations.bit_resize(self.zero_one_bin, 5, resize_flags=aerospike.BIT_RESIZE_FROM_FRONT)]
        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)

        assert len(bins[self.zero_one_bin]) == 5
        assert bins[self.zero_one_bin] == bytearray([1] * 5)

    def test_bit_resize_default_allows_create(self):
        """
        By default we can create a new bin with resize.
        """
        ops = [bitwise_operations.bit_resize("new_bin_name", 10)]
        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        assert bins["new_bin_name"] == bytearray([0] * 10)

    def test_bit_resize_create_only_allows_create(self):
        """
        Create a bin with resize using the create only flag.
        """
        bit_policy = {"bit_write_flags": aerospike.BIT_WRITE_CREATE_ONLY}
        ops = [bitwise_operations.bit_resize("new_bin_name", 10, policy=bit_policy)]
        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        assert bins["new_bin_name"] == bytearray([0] * 10)

    def test_bit_resize_update_only_allows_update(self):
        """
        Update a bin with resize using the update only flag.
        """
        bit_policy = {"bit_write_flags": aerospike.BIT_WRITE_UPDATE_ONLY}
        ops = [bitwise_operations.bit_resize(self.test_bin_zeroes, 10, policy=bit_policy)]
        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        assert bins[self.test_bin_zeroes] == bytearray([0] * 10)

    def test_bit_resize_create_only_prevents_update(self):
        """
        Attempt to update a bin using the create only flag, should fail.
        """
        bit_policy = {"bit_write_flags": aerospike.BIT_WRITE_CREATE_ONLY}
        ops = [bitwise_operations.bit_resize(self.test_bin_zeroes, 10, policy=bit_policy)]
        with pytest.raises(e.BinExistsError):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_resize_update_only_prevents_create(self):
        """
        Attempt to create a bin with the update only flag.
        """
        bit_policy = {"bit_write_flags": aerospike.BIT_WRITE_UPDATE_ONLY}
        ops = [bitwise_operations.bit_resize("new_binname", 10, policy=bit_policy)]
        with pytest.raises(e.BinNotFound):
            self.as_connection.operate(self.test_key, ops)

    # TODO: duplicate test name
    @pytest.mark.xfail(reason="Had a duplicate test name, but fails when renamed.")
    def test_bit_resize_partial_no_fail_duplicate(self):
        """
        By default we can create a new bin with resize.
        """
        bit_policy = {
            "bit_write_flags": aerospike.BIT_WRITE_CREATE_ONLY
            | aerospike.BIT_WRITE_NO_FAIL
            | aerospike.BIT_WRITE_PARTIAL
        }
        ops = [
            bitwise_operations.bit_resize(self.test_bin_zeroes, 15, policy=bit_policy),
            bitwise_operations.bit_resize(self.test_bin_zeroes, 20),
        ]
        self.as_connection.operate(self.test_key, ops)
        _, _, bins = self.as_connection.get(self.test_key)
        assert bins[self.test_bin_zeroes] == bytearray([0] * 20)

    def test_bit_resize_partial_no_fail(self):
        """
        By default we can create a new bin with resize.
        """
        bit_policy = {"bit_write_flags": aerospike.BIT_WRITE_UPDATE_ONLY | aerospike.BIT_WRITE_NO_FAIL}
        ops = [bitwise_operations.bit_resize("new_binname", 10, policy=bit_policy)]
        self.as_connection.operate(self.test_key, ops)
        _, _, bins = self.as_connection.get(self.test_key)
        assert "new_binname" not in bins

    def test_bit_resize_grow_only_does_not_allow_shrink(self):
        """
        Resize with grow only flags.
        """
        ops = [bitwise_operations.bit_resize(self.test_bin_ones, 1, resize_flags=aerospike.BIT_RESIZE_GROW_ONLY)]
        with pytest.raises(e.InvalidRequest):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_resize_shrink_only_does_not_allow_grow(self):
        """
        Prevent bin growth with shrin only flag.
        """
        ops = [bitwise_operations.bit_resize(self.test_bin_ones, 10, resize_flags=aerospike.BIT_RESIZE_SHRINK_ONLY)]
        with pytest.raises(e.InvalidRequest):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_remove_two_bytes(self):
        """
        Perform a bit_remove op with offset 1 and byte_size 2.
        """
        ops = [bitwise_operations.bit_remove(self.test_bin_zeroes, 1, 2, None)]
        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        assert bins[self.test_bin_zeroes] == bytearray([0] * 3)
        # should have removed 2 0s)

        assert len(bins[self.test_bin_zeroes]) == 3
        # should have len 3 after removing 2 0s

    def test_bit_remove_randnumbytes_randoffset(self):
        """
        Perform a bit_remove op with random offset and random byte_size.
        """
        offset = random.randint(0, 4)
        num_bytes = random.randint(1, (5 - offset))
        ops = [bitwise_operations.bit_remove(self.test_bin_zeroes, offset, num_bytes, None)]
        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        assert bins[self.test_bin_zeroes] == bytearray([0] * (5 - num_bytes))
        # should have removed num_bytes 0s

    def test_bit_remove_offset_out_of_range(self):
        """
        Perform a bit_remove op with an offset outside the bit_map being
        modified.
        """
        ops = [bitwise_operations.bit_remove(self.test_bin_zeroes, 6, 1, None)]

        with pytest.raises(e.InvalidRequest):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_remove_byte_size_too_large(self):
        """
        Perform a bit_remove op with byte_size larger than bit_map being modified.
        """
        ops = [bitwise_operations.bit_remove(self.test_bin_zeroes, 0, 6, None)]

        with pytest.raises(e.InvalidRequest):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_remove_byte_offset_with_byte_too_large(self):
        """
        Perform a bit_remove op with an offset and byte_size that attempt
        to modify the bitmap outside of bounds.
        """
        ops = [bitwise_operations.bit_remove(self.test_bin_zeroes, 3, 3, None)]

        with pytest.raises(e.InvalidRequest):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_remove_bad_bin_name(self):
        """
        Perform a bit_remove op with on a non existent bin.
        """
        ops = [bitwise_operations.bit_remove("bad_name", 3, 3, None)]

        with pytest.raises(e.BinNotFound):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_remove_bad_arg_type(self):
        """
        Perform a bit_remove op with an unsupported float type argument.
        """
        ops = [bitwise_operations.bit_remove("bad_name", 3, 3.5, None)]

        with pytest.raises(e.ParamError):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_set_bit_random_byte_random_offset(self):
        """
        Perform a bit_set op with a random offset and random valued byte.
        """
        value = bytearray()
        rand_byte = random.randint(0, 255)
        value.append(rand_byte)
        rand_offset = random.randint(0, 4) * 8
        ops = [bitwise_operations.bit_set(self.test_bin_zeroes, rand_offset, 8, 1, value, None)]
        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([0] * 5)
        expected_result[rand_offset // 8] = rand_byte
        assert bins[self.test_bin_zeroes] == expected_result
        # should set the byte at rand_offset/8 to rand_byte

    def test_bit_set_bit_inbetween_bytes(self):
        """
        Perform a bit_set op with an offset that places the value across bytes.
        """
        value = bytearray()
        value.append(255)
        ops = [bitwise_operations.bit_set(self.test_bin_zeroes, 4, 8, 1, value, None)]
        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([15] * 1 + [240] * 1 + [0] * 3)
        assert bins[self.test_bin_zeroes] == expected_result

    def test_bit_set_bit_multiple_bytes(self):
        """
        Perform a bit_set op with a random number of bytes.
        """
        value = bytearray()
        value = bytearray()
        rand_byte = random.randint(0, 255)
        num_bytes = random.randint(1, 5)
        for x in range(0, num_bytes):
            value.append(rand_byte)
        ops = [bitwise_operations.bit_set(self.test_bin_zeroes, 0, (num_bytes * 8), num_bytes, value, None)]
        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray(([rand_byte] * num_bytes) + [0] * (5 - num_bytes))
        assert bins[self.test_bin_zeroes] == expected_result

    def test_bit_set_bit_index_out_of_range(self):
        """'
        Perform a bit_set op with a bit offset that is larger than the bit_map being modified.
        """
        value = bytearray()
        value.append(255)
        ops = [bitwise_operations.bit_set(self.test_bin_zeroes, 41, 8, 1, value, None)]
        with pytest.raises(e.OpNotApplicable):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_set_bit_value_size_too_large(self):
        """
        Perform a bit_set op with a value larger than the bit_map being modified.
        """
        value = bytearray()
        for x in range(0, 5):
            value.append(255)
        ops = [bitwise_operations.bit_set(self.test_bin_zeroes, 0, 48, 6, value, None)]
        with pytest.raises(e.OpNotApplicable):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_set_bit_bit_size_too_large(self):
        """
        Perform a bit_set op with an bit size larger than the actual value.
        """
        value = bytearray()
        value.append(255)
        ops = [bitwise_operations.bit_set(self.test_bin_zeroes, 0, 16, 1, value, None)]
        with pytest.raises(e.InvalidRequest):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_set_bit_invalid_arg_type(self):
        """
        Perform a bit_set op with an unsupported float.
        """
        value = 85323.9
        ops = [bitwise_operations.bit_set(self.test_bin_zeroes, 0, 8, 1, value, None)]
        with pytest.raises(e.ParamError):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_set_bit_bad_bin_name(self):
        """
        Perform a bit_set op with a non existent bin.
        """
        value = bytearray()
        value.append(255)
        ops = [bitwise_operations.bit_set("bad_name", 0, 8, 1, value, None)]
        with pytest.raises(e.BinNotFound):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_count_seven(self):
        """
        Perform a bitwise count op.
        """
        ops = [bitwise_operations.bit_count(self.count_bin, 20, 9)]

        _, _, result = self.as_connection.operate(self.test_key, ops)
        assert result["count"] == 7

    def test_bit_count_one(self):
        """
        Perform a bitwise count op.
        """
        ops = [bitwise_operations.bit_count(self.zero_one_bin, 47, 8)]

        _, _, result = self.as_connection.operate(self.test_key, ops)
        assert result["bitwise01"] == 1

    def test_bit_count_random_bit_size(self):
        """
        Perform a bitwise count op.
        """
        bit_size = random.randint(1, 40)
        ops = [bitwise_operations.bit_count(self.five_255_bin, 0, bit_size)]

        _, _, result = self.as_connection.operate(self.test_key, ops)
        assert result["255"] == bit_size

    def test_bit_count_bit_offset_out_of_range(self):
        """
        Attempts to perform a bitwise count op with bit_offset outside of
        the bitmap being counted.
        """
        ops = [bitwise_operations.bit_count(self.zero_one_bin, 81, 1)]

        with pytest.raises(e.OpNotApplicable):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_count_bit_size_too_large(self):
        """
        Attempts to perform a bitwise count op on more bits than exist.
        """
        ops = [bitwise_operations.bit_count(self.zero_one_bin, 1, 81)]

        with pytest.raises(e.OpNotApplicable):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_count_bit_size_with_offset_too_large(self):
        """
        Attempts to perform a bitwise count past the bounds of the bitmap being counted.
        """
        ops = [bitwise_operations.bit_count(self.zero_one_bin, 40, 41)]

        with pytest.raises(e.OpNotApplicable):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_add_simple(self):
        """
        Perform a bitwise add op.
        """
        ops = [bitwise_operations.bit_add(self.test_bin_ones, 0, 8, 1, False, aerospike.BIT_OVERFLOW_FAIL, None)]

        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([2] * 1 + [1] * 4)
        assert bins[self.test_bin_ones] == expected_result

    def test_bit_add_inbetween_bytes(self):
        """
        Perform a bitwise add op with an offset that lands inbetween bytes.
        """
        ops = [bitwise_operations.bit_add(self.test_bin_zeroes, 4, 8, 255, False, aerospike.BIT_OVERFLOW_FAIL, None)]

        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([15] * 1 + [240] * 1 + [0] * 3)
        assert bins[self.test_bin_zeroes] == expected_result

    def test_bit_add_multiple_bytes(self):
        """
        Perform a bitwise add op with multiple bytes.
        """
        ops = [bitwise_operations.bit_add(self.test_bin_zeroes, 8, 16, 65535, False, aerospike.BIT_OVERFLOW_FAIL, None)]

        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([0] * 1 + [255] * 2 + [0] * 2)
        assert bins[self.test_bin_zeroes] == expected_result

    def test_bit_add_bad_bin_name(self):
        """
        Perform a bitwise add op on a nonexistent bin.
        """
        ops = [bitwise_operations.bit_add("bad_name", 8, 16, 65535, False, aerospike.BIT_OVERFLOW_FAIL, None)]

        with pytest.raises(e.BinNotFound):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_add_bit_offset_out_of_range(self):
        """
        Perform a bitwise add op with a bit_offset that is out of range
        for the bitmap being modified.
        """
        ops = [bitwise_operations.bit_add(self.test_bin_zeroes, 41, 1, 1, False, aerospike.BIT_OVERFLOW_FAIL, None)]

        with pytest.raises(e.OpNotApplicable):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_add_bit_size_out_of_range(self):
        """
        Perform a bitwise add op with a bit size too large for
        the bitmap being modified.
        """
        ops = [bitwise_operations.bit_add(self.test_bin_zeroes, 0, 41, 1, False, aerospike.BIT_OVERFLOW_FAIL, None)]

        with pytest.raises(e.OpNotApplicable):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_add_bit_size_signed(self):
        """
        Perform a bitwise add op subtraction.
        """
        ops = [bitwise_operations.bit_add(self.five_255_bin, 0, 8, 254, True, aerospike.BIT_OVERFLOW_WRAP, None)]

        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([253] * 1 + [255] * 4)
        assert bins[self.five_255_bin] == expected_result

    def test_bit_add_overflow_fail(self):
        """
        Perform a bitwise add op that overflows with the BIT_OVERFLOW_FAIL action.
        """
        ops = [bitwise_operations.bit_add(self.five_255_bin, 0, 8, 1, False, aerospike.BIT_OVERFLOW_FAIL, None)]

        with pytest.raises(e.OpNotApplicable):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_add_overflow_saturate(self):
        """
        Perform a bitwise add op that overflows with the BIT_OVERFLOW_SATURATE action.
        """
        ops = [bitwise_operations.bit_add(self.five_255_bin, 0, 8, 1, False, aerospike.BIT_OVERFLOW_SATURATE, None)]

        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([255] * 5)
        assert bins[self.five_255_bin] == expected_result

    def test_bit_add_overflow_wrap(self):
        """
        Perform a bitwise add op that overflows with the BIT_OVERFLOW_WRAP action.
        """
        ops = [bitwise_operations.bit_add(self.five_255_bin, 0, 8, 1, False, aerospike.BIT_OVERFLOW_WRAP, None)]

        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([0] * 1 + [255] * 4)
        assert bins[self.five_255_bin] == expected_result

    def test_bit_and(self):
        """
        Perform a bitwise and op.
        """
        value = bytearray()
        value.append(0)
        ops = [bitwise_operations.bit_and(self.five_255_bin, 0, 8, 1, value, None)]

        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([0] * 1 + [255] * 4)
        assert bins[self.five_255_bin] == expected_result

    def test_bit_and_across_bytes(self):
        """
        Perform a bitwise and op with a bit_offset that causes the op to span bytes.
        """
        value = bytearray()
        value.append(0)
        ops = [bitwise_operations.bit_and(self.five_255_bin, 7, 8, 1, value, None)]

        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([254] * 1 + [1] * 1 + [255] * 3)
        assert bins[self.five_255_bin] == expected_result

    def test_bit_and_multiple_bytes(self):
        """
        Perform a bitwise and op with value large enough to span multiple bytes.
        """
        value = bytearray()
        value.append(1)
        value.append(1)
        value.append(1)
        ops = [bitwise_operations.bit_and(self.five_255_bin, 8, 17, 3, value, None)]

        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([255] * 1 + [1] * 2 + [127] * 1 + [255] * 1)
        assert bins[self.five_255_bin] == expected_result

    def test_bit_and_offset_out_of_range(self):
        """
        Perform a bitwise and op with a bit_offset outside of the bitmap being modified.
        """
        value = bytearray()
        value.append(0)
        ops = [bitwise_operations.bit_and(self.five_255_bin, 41, 8, 1, value, None)]

        with pytest.raises(e.OpNotApplicable):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_and_offset_bit_size_larger_than_val(self):
        """
        Perform a bitwise and op with a bit_size > the size of value.
        """
        value = bytearray()
        value.append(0)
        ops = [bitwise_operations.bit_and(self.five_255_bin, 0, 16, 1, value, None)]

        with pytest.raises(e.InvalidRequest):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_and_offset_value_byte_size_too_large(self):
        """
        Perform a bitwise and op with a value_byte_size larger than the bitmap
        being modified.
        """
        value = bytearray()
        for x in range(0, 5):
            value.append(0)
        ops = [bitwise_operations.bit_and(self.five_255_bin, 0, 48, 6, value, None)]

        with pytest.raises(e.OpNotApplicable):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_and_invalid_bin_name(self):
        """
        Perform a bitwise and op with a non existent bin name.
        """
        value = bytearray()
        value.append(0)
        ops = [bitwise_operations.bit_and("bad_name", 0, 8, 1, value, None)]

        with pytest.raises(e.BinNotFound):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_and_invalid_value(self):
        """
        Perform a bitwise and op with a non existent bin name.
        """
        ops = [bitwise_operations.bit_and("bad_name", 0, 8, 1, 1.5, None)]

        with pytest.raises(e.ParamError):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_get(self):
        """
        Perform a bitwise get op.
        """
        ops = [bitwise_operations.bit_get(self.five_255_bin, 0, 8)]

        _, _, result = self.as_connection.operate(self.test_key, ops)

        expected_result = bytearray([255] * 1)
        assert result["255"] == expected_result

    def test_bit_get_negative_offset(self):
        """
        Perform a bitwise get op with a negative offset.
        """
        ops = [bitwise_operations.bit_get(self.count_bin, -2, 2)]

        _, _, result = self.as_connection.operate(self.test_key, ops)

        expected_result = bytearray([192] * 1)
        assert result[self.count_bin] == expected_result

    def test_bit_get_accross_bytes(self):
        """
        Perform a bitwise get op across bytes.
        """
        ops = [bitwise_operations.bit_get(self.test_bin_ones, 4, 8)]

        _, _, result = self.as_connection.operate(self.test_key, ops)

        expected_result = bytearray([16] * 1)
        assert result["bitwise1"] == expected_result

    def test_bit_get_fraction_of_byte(self):
        """
        Perform a bitwise get op on a portion of a byte.
        """
        ops = [bitwise_operations.bit_get(self.five_255_bin, 4, 2)]

        _, _, result = self.as_connection.operate(self.test_key, ops)

        expected_result = bytearray([192] * 1)
        assert result["255"] == expected_result

    def test_bit_get_multiple_bytes(self):
        """
        Perform a bitwise get op across multiple bytes.
        """
        ops = [bitwise_operations.bit_get(self.five_255_bin, 4, 17)]

        _, _, result = self.as_connection.operate(self.test_key, ops)

        expected_result = bytearray([255] * 2 + [128] * 1)
        assert result["255"] == expected_result

    def test_bit_get_bit_offset_out_of_range(self):
        """
        Perform a bitwise get with a bitoffset outside of the bitmap
        being read.
        """
        ops = [bitwise_operations.bit_get(self.test_bin_ones, 41, 1)]

        with pytest.raises(e.OpNotApplicable):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_get_bit_size_too_large(self):
        """
        Perform a bitwise get with a bit_size larger than the bitmap
        being read.
        """
        ops = [bitwise_operations.bit_get(self.test_bin_ones, 0, 41)]

        with pytest.raises(e.OpNotApplicable):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_get_bad_bin_name(self):
        """
        Perform a bitwise get with a non existent bin.
        """
        ops = [bitwise_operations.bit_get("bad_name", 0, 1)]
        _, _, result = self.as_connection.operate(self.test_key, ops)

        expected_result = None
        assert result["bad_name"] == expected_result

    def test_bit_get_bad_argument_type(self):
        """
        Perform a bitwise get with a float for bit_size.
        """
        ops = [bitwise_operations.bit_get(self.test_bin_ones, 0, 1.5)]

        with pytest.raises(e.ParamError):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_get_int(self):
        """
        Perform a bitwise get int op.
        """
        ops = [bitwise_operations.bit_get_int(self.five_255_bin, 0, 8, False)]

        _, _, result = self.as_connection.operate(self.test_key, ops)

        expected_result = 255
        assert result["255"] == expected_result

    def test_bit_get_int_accross_bytes(self):
        """
        Perform a bitwise get int op across bytes.
        """
        ops = [bitwise_operations.bit_get_int(self.test_bin_ones, 4, 8, False)]

        _, _, result = self.as_connection.operate(self.test_key, ops)

        expected_result = 16
        assert result["bitwise1"] == expected_result

    def test_bit_get_int_fraction_of_byte(self):
        """
        Perform a bitwise get op on a portion of a byte.
        """
        ops = [bitwise_operations.bit_get_int(self.five_255_bin, 4, 2, False)]

        _, _, result = self.as_connection.operate(self.test_key, ops)

        expected_result = 3
        assert result["255"] == expected_result

    def test_bit_get_int_multiple_bytes(self):
        """
        Perform a bitwise get int op across multiple bytes.
        """
        ops = [bitwise_operations.bit_get_int(self.five_255_bin, 4, 17, False)]

        _, _, result = self.as_connection.operate(self.test_key, ops)

        expected_result = 131071
        assert result["255"] == expected_result

    def test_bit_get_int_bit_offset_out_of_range(self):
        """
        Perform a bitwise get int with a bitoffset outside of the bitmap
        being read.
        """
        ops = [bitwise_operations.bit_get_int(self.test_bin_ones, 41, 1, False)]

        with pytest.raises(e.OpNotApplicable):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_get_int_bit_size_too_large(self):
        """
        Perform a bitwise get int with a bit_size larger than the bitmap
        being read.
        """
        ops = [bitwise_operations.bit_get_int(self.test_bin_ones, 0, 41, False)]

        with pytest.raises(e.OpNotApplicable):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_get_int_bad_bin_name(self):
        """
        Perform a bitwise get int with a non existent bin.
        """
        ops = [bitwise_operations.bit_get_int("bad_name", 0, 1, False)]
        _, _, result = self.as_connection.operate(self.test_key, ops)

        expected_result = None
        assert result["bad_name"] == expected_result

    def test_bit_get_int_bad_argument_type(self):
        """
        Perform a bitwise get int with a float for bit_size.
        """
        ops = [bitwise_operations.bit_get_int(self.test_bin_ones, 0, 1.5, False)]

        with pytest.raises(e.ParamError):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_get_int_signed(self):
        """
        Perform a bitwise get int op with sign true.
        """
        ops = [bitwise_operations.bit_get_int(self.five_255_bin, 0, 8, True)]

        _, _, result = self.as_connection.operate(self.test_key, ops)

        expected_result = -1
        assert result["255"] == expected_result

    def test_bit_insert(self):
        """
        Perform a bitwise insert op.
        """
        value = bytearray([3])
        ops = [bitwise_operations.bit_insert(self.test_bin_zeroes, 0, 1, value, None)]
        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([3] * 1 + [0] * 5)
        assert bins[self.test_bin_zeroes] == expected_result

    def test_bit_insert_multiple_bytes(self):
        """
        Perform a bitwise insert op.
        """
        value = bytearray([3] * 3)
        ops = [bitwise_operations.bit_insert(self.test_bin_zeroes, 0, 3, value, None)]
        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([3] * 3 + [0] * 5)
        assert bins[self.test_bin_zeroes] == expected_result

    def test_bit_insert_multiple_bytes_with_offset(self):
        """
        Perform a bitwise insert op with multiple bytes and a non 0 offset.
        """
        value = bytearray([3] * 3)
        ops = [bitwise_operations.bit_insert(self.test_bin_zeroes, 2, 3, value, None)]
        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([0] * 2 + [3] * 3 + [0] * 3)
        assert bins[self.test_bin_zeroes] == expected_result

    def test_bit_insert_offset_out_of_range(self):
        """
        Perform a bitwise insert op where byte_offset is out of range for
        the bitmap being modified. Places 0 untill proper offset is reached.
        """
        value = bytearray([3])
        ops = [bitwise_operations.bit_insert(self.five_255_bin, 9, 1, value, None)]

        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([255] * 5 + [0] * 4 + [3] * 1)
        assert bins[self.five_255_bin] == expected_result

    def test_bit_insert_value_byte_size_too_large(self):
        """
        Perform a bitwise insert op where value_byte_size is larger than the bitmap
        being modified.
        """
        value = bytearray([3] * 6)
        ops = [bitwise_operations.bit_insert(self.five_255_bin, 0, 6, value, None)]

        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([3] * 6 + [255] * 5)
        assert bins[self.five_255_bin] == expected_result

    def test_bit_insert_value_byte_size_smaller_than_value(self):
        """
        Perform a bitwise insert op where value_byte_size is smaller than the bitmap
        being modified.
        """
        value = bytearray([3] * 6)
        ops = [bitwise_operations.bit_insert(self.five_255_bin, 0, 2, value, None)]

        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([3] * 2 + [255] * 5)
        assert bins[self.five_255_bin] == expected_result

    def test_bit_insert_nonexistent_bin_name(self):
        """
        Perform a bitwise insert op with a non existent bin.
        """
        value = bytearray([3])
        ops = [bitwise_operations.bit_insert("bad_name", 0, 1, value, None)]

        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([3])
        assert bins["bad_name"] == expected_result

    def test_bit_insert_bad_arg_type(self):
        """
        Perform a bitwise insert op with a float byte_size.
        """
        value = bytearray([3])
        ops = [bitwise_operations.bit_insert(self.five_255_bin, 0, 1.5, value, None)]

        with pytest.raises(e.ParamError):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_lscan(self):
        """
        Perform a bitwise lscan op.
        """
        value = True
        ops = [bitwise_operations.bit_lscan(self.count_bin, 32, 8, value)]

        expected_value = 6
        _, _, result = self.as_connection.operate(self.test_key, ops)
        assert result[self.count_bin] == expected_value

    def test_bit_lscan_across_bytes(self):
        """
        Perform a bitwise lscan op with an offset that causes the scan to
        search multiple bytes.
        """
        value = False
        ops = [bitwise_operations.bit_lscan(self.test_bin_ones, 7, 8, value)]

        expected_value = 1
        _, _, result = self.as_connection.operate(self.test_key, ops)
        assert result[self.test_bin_ones] == expected_value

    def test_bit_lscan_value_not_found(self):
        """
        Perform a bitwise lscan op with an offset that causes the scan to
        search multiple bytes.
        """
        value = False
        ops = [bitwise_operations.bit_lscan(self.five_255_bin, 0, 40, value)]

        expected_value = -1
        _, _, result = self.as_connection.operate(self.test_key, ops)
        assert result[self.five_255_bin] == expected_value

    def test_bit_lscan_offset_out_of_range(self):
        """
        Perform a bitwise lscan op with bit_offset outside bitmap.S
        """
        value = True
        ops = [bitwise_operations.bit_lscan(self.count_bin, 41, 8, value)]

        with pytest.raises(e.OpNotApplicable):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_lscan_bit_size_too_large(self):
        """
        Perform a bitwise lscan op with bit_size larger than bitmap.S
        """
        value = True
        ops = [bitwise_operations.bit_lscan(self.test_bin_ones, 0, 41, value)]

        with pytest.raises(e.OpNotApplicable):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_lscan_bad_bin_name(self):
        """
        Perform a bitwise lscan op with a non existent bin.
        """
        value = True
        ops = [bitwise_operations.bit_lscan("bad_name", 0, 8, value)]

        expected_value = None
        _, _, result = self.as_connection.operate(self.test_key, ops)
        assert result["bad_name"] == expected_value

    def test_bit_lshift(self):
        """
        Perform a bitwise lshift op.
        """
        ops = [bitwise_operations.bit_lshift(self.test_bin_ones, 0, 8, 3, None)]

        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([8] * 1 + [1] * 4)
        assert bins[self.test_bin_ones] == expected_result

    def test_bit_lshift_across_bytes(self):
        """
        Perform a bitwise lshift op with bit_offset causing the op
        to cross bytes.
        """
        ops = [bitwise_operations.bit_lshift(self.test_bin_ones, 4, 12, 3, None)]

        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([8] * 2 + [1] * 3)
        assert bins[self.test_bin_ones] == expected_result

    def test_bit_lshift_wrap(self):
        """
        Perform a bitwise lshift op with shift > 1 byte.
        """
        ops = [bitwise_operations.bit_lshift(self.test_bin_ones, 0, 40, 8, None)]

        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([1] * 4 + [0])
        assert bins[self.test_bin_ones] == expected_result

    def test_bit_lshift_offset_out_of_range(self):
        """
        Perform a bitwise lshift op with offset > bitmap.
        """
        ops = [bitwise_operations.bit_lshift(self.test_bin_ones, 41, 8, 1, None)]

        with pytest.raises(e.OpNotApplicable):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_lshift_bit_size_too_large(self):
        """
        Perform a bitwise lshift op with bit_size > bitmap.
        """
        ops = [bitwise_operations.bit_lshift(self.test_bin_ones, 0, 41, 1, None)]

        with pytest.raises(e.OpNotApplicable):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_lshift_bad_bin_name(self):
        """
        Perform a bitwise lshift op with non existent bin.
        """
        ops = [bitwise_operations.bit_lshift("bad_name", 0, 8, 1, None)]

        with pytest.raises(e.BinNotFound):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_lshift_bad_arg(self):
        """
        Perform a bitwise lshift op with a float for bit_offset.
        """
        ops = [bitwise_operations.bit_lshift(self.test_bin_ones, 1.5, 8, 1, None)]

        with pytest.raises(e.ParamError):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_not(self):
        """
        Perform a bitwise not op.
        """
        ops = [bitwise_operations.bit_not(self.five_255_bin, 0, 40, None)]

        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([0] * 5)
        assert bins[self.five_255_bin] == expected_result

    def test_bit_not_small_bit_size(self):
        """
        Perform a bitwise not op with a bit size < 1 byte.
        """
        ops = [bitwise_operations.bit_not(self.five_255_bin, 5, 3, None)]

        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([248] + [255] * 4)
        assert bins[self.five_255_bin] == expected_result

    def test_bit_not_offset_out_of_range(self):
        """
        Perform a bitwise not op with a bit_offset outside the bitmap
        being modified.
        """
        ops = [bitwise_operations.bit_not(self.five_255_bin, 41, 8, None)]

        with pytest.raises(e.OpNotApplicable):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_not_bit_size_too_large(self):
        """
        Perform a bitwise not op with a bit_size > bitmap.
        """
        ops = [bitwise_operations.bit_not(self.five_255_bin, 0, 41, None)]

        with pytest.raises(e.OpNotApplicable):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_not_bad_bin_name(self):
        """
        Perform a bitwise not op with a bit_size > bitmap.
        """
        ops = [bitwise_operations.bit_not("bad_name", 0, 8, None)]

        with pytest.raises(e.BinNotFound):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_not_bad_arg(self):
        """
        Perform a bitwise not op with a float for bit_offset.
        """
        ops = [bitwise_operations.bit_not(self.five_255_bin, 2.7, 8, None)]

        with pytest.raises(e.ParamError):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_or(self):
        """
        Perform a bitwise or op.
        """
        value = bytearray()
        value.append(8)
        ops = [bitwise_operations.bit_or(self.test_bin_ones, 0, 8, 1, value, None)]

        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([9] * 1 + [1] * 4)
        assert bins[self.test_bin_ones] == expected_result

    def test_bit_or_multiple_bytes(self):
        """
        Perform a bitwise or op on multiple bytes.
        """
        value = bytearray([8] * 5)
        ops = [bitwise_operations.bit_or(self.test_bin_ones, 0, 40, 5, value, None)]

        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([9] * 5)
        assert bins[self.test_bin_ones] == expected_result

    def test_bit_or_multiple_bytes_value_unchanged(self):
        """
        Perform a bitwise or op.
        """
        value = bytearray([255])
        ops = [bitwise_operations.bit_or(self.five_255_bin, 7, 8, 1, value, None)]

        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([255] * 5)
        assert bins[self.five_255_bin] == expected_result

    def test_bit_or_bit_offset_out_of_range(self):
        """
        Perform a bitwise or op with bit_offset > bitmap.
        """
        value = bytearray()
        value.append(8)
        ops = [bitwise_operations.bit_or(self.test_bin_ones, 41, 8, 1, value, None)]

        with pytest.raises(e.OpNotApplicable):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_or_bit_size_larger_than_value(self):
        """
        Perform a bitwise or op with bit_size > value.
        """
        value = bytearray()
        value.append(8)
        ops = [bitwise_operations.bit_or(self.test_bin_ones, 0, 9, 1, value, None)]

        with pytest.raises(e.InvalidRequest):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_or_bit_size_too_large(self):
        """
        Perform a bitwise or op with bit_size > bitmap.
        """
        value = bytearray([8] * 6)
        ops = [bitwise_operations.bit_or(self.test_bin_ones, 0, 41, 6, value, None)]

        with pytest.raises(e.OpNotApplicable):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_or_bad_bin_name(self):
        """
        Perform a bitwise or op with a non existent bin.
        """
        value = bytearray([8])
        ops = [bitwise_operations.bit_or("bad_name", 0, 8, 1, value, None)]

        with pytest.raises(e.BinNotFound):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_or_bad_arg(self):
        """
        Perform a bitwise or op with an integer for value.
        """
        value = 1
        ops = [bitwise_operations.bit_or("bad_name", 0, 8, 1, value, None)]

        with pytest.raises(e.ParamError):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_rscan(self):
        """
        Perform a bitwise rscan op.
        """
        value = True
        ops = [bitwise_operations.bit_rscan(self.count_bin, 32, 8, value)]

        expected_value = 7
        _, _, result = self.as_connection.operate(self.test_key, ops)
        assert result[self.count_bin] == expected_value

    def test_bit_rscan_across_bytes(self):
        """
        Perform a bitwise rscan op with an offset that causes the scan to
        search multiple bytes.
        """
        value = False
        ops = [bitwise_operations.bit_rscan(self.count_bin, 4, 8, value)]

        expected_value = 6
        _, _, result = self.as_connection.operate(self.test_key, ops)
        assert result[self.count_bin] == expected_value

    def test_bit_rscan_value_not_found(self):
        """
        Perform a bitwise rscan op with an offset that causes the scan to
        search multiple bytes.
        """
        value = False
        ops = [bitwise_operations.bit_rscan(self.five_255_bin, 0, 40, value)]

        expected_value = -1
        _, _, result = self.as_connection.operate(self.test_key, ops)
        assert result[self.five_255_bin] == expected_value

    def test_bit_rscan_offset_out_of_range(self):
        """
        Perform a bitwise rscan op with bit_offset outside bitmap.
        """
        value = True
        ops = [bitwise_operations.bit_rscan(self.count_bin, 41, 8, value)]

        with pytest.raises(e.OpNotApplicable):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_rscan_bit_size_too_large(self):
        """
        Perform a bitwise rscan op with bit_size larger than bitmap.
        """
        value = True
        ops = [bitwise_operations.bit_rscan(self.test_bin_ones, 0, 41, value)]

        with pytest.raises(e.OpNotApplicable):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_rscan_bad_bin_name(self):
        """
        Perform a bitwise rscan op with a non existent bin.
        """
        value = True
        ops = [bitwise_operations.bit_rscan("bad_name", 0, 8, value)]

        expected_value = None
        _, _, result = self.as_connection.operate(self.test_key, ops)
        assert result["bad_name"] == expected_value

    def test_bit_rshift(self):
        """
        Perform a bitwise rshift op.
        """
        ops = [bitwise_operations.bit_rshift(self.count_bin, 8, 8, 3, None)]

        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([1] * 1 + [10] * 1 + [86] * 1 + [255] * 1 + [3] * 1)
        assert bins[self.count_bin] == expected_result

    def test_bit_rshift_across_bytes(self):
        """
        Perform a bitwise rshift op with bit_offset causing the op
        to cross bytes.
        """
        ops = [bitwise_operations.bit_rshift(self.test_bin_ones, 4, 16, 3, None)]

        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([0] + [32] + [33] + [1] * 2)
        assert bins[self.test_bin_ones] == expected_result

    def test_bit_rshift_wrap(self):
        """
        Perform a bitwise rshift op with shift > 1 byte.
        """
        ops = [bitwise_operations.bit_rshift(self.test_bin_ones, 0, 40, 8, None)]

        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([0] + [1] * 4)
        assert bins[self.test_bin_ones] == expected_result

    def test_bit_rshift_offset_out_of_range(self):
        """
        Perform a bitwise rshift op with offset > bitmap.
        """
        ops = [bitwise_operations.bit_rshift(self.test_bin_ones, 41, 8, 1, None)]

        with pytest.raises(e.OpNotApplicable):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_rshift_bit_size_too_large(self):
        """
        Perform a bitwise rshift op with bit_size > bitmap.
        """
        ops = [bitwise_operations.bit_rshift(self.test_bin_ones, 0, 41, 1, None)]

        with pytest.raises(e.OpNotApplicable):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_rshift_bad_bin_name(self):
        """
        Perform a bitwise rshift op with non existent bin.
        """
        ops = [bitwise_operations.bit_rshift("bad_name", 0, 8, 1, None)]

        with pytest.raises(e.BinNotFound):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_rshift_bad_arg(self):
        """
        Perform a bitwise rshift op with a float for bit_offset.
        """
        ops = [bitwise_operations.bit_rshift(self.test_bin_ones, 1.5, 8, 1, None)]

        with pytest.raises(e.ParamError):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_subtract_simple(self):
        """
        Perform a bitwise subtract op.
        """
        ops = [bitwise_operations.bit_subtract(self.test_bin_ones, 0, 8, 1, False, aerospike.BIT_OVERFLOW_FAIL, None)]

        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([0] * 1 + [1] * 4)
        assert bins[self.test_bin_ones] == expected_result

    def test_bit_subtract_inbetween_bytes(self):
        """
        Perform a bitwise subtract op with an offset that lands inbetween bytes.
        """
        ops = [bitwise_operations.bit_subtract(self.five_255_bin, 4, 8, 255, False, aerospike.BIT_OVERFLOW_FAIL, None)]

        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([240] * 1 + [15] * 1 + [255] * 3)
        assert bins[self.five_255_bin] == expected_result

    def test_bit_subtract_multiple_bytes(self):
        """
        Perform a bitwise subtract op with multiple bytes.
        """
        ops = [
            bitwise_operations.bit_subtract(self.test_bin_ones, 8, 16, 257, False, aerospike.BIT_OVERFLOW_FAIL, None)
        ]

        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([1] * 1 + [0] * 2 + [1] * 2)
        assert bins[self.test_bin_ones] == expected_result

    def test_bit_subtract_bad_bin_name(self):
        """
        Perform a bitwise subtract op on a nonexistent bin.
        """
        ops = [bitwise_operations.bit_subtract("bad_name", 8, 16, 257, False, aerospike.BIT_OVERFLOW_FAIL, None)]

        with pytest.raises(e.BinNotFound):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_subtract_bit_offset_out_of_range(self):
        """
        Perform a bitwise subtract op with a bit_offset that is out of range
        for the bitmap being modified.
        """
        ops = [
            bitwise_operations.bit_subtract(self.test_bin_zeroes, 41, 1, 1, False, aerospike.BIT_OVERFLOW_FAIL, None)
        ]

        with pytest.raises(e.OpNotApplicable):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_subtract_bit_size_out_of_range(self):
        """
        Perform a bitwise subtract op with a bit size too large for
        the bitmap being modified.
        """
        ops = [
            bitwise_operations.bit_subtract(self.test_bin_zeroes, 0, 41, 1, False, aerospike.BIT_OVERFLOW_FAIL, None)
        ]

        with pytest.raises(e.OpNotApplicable):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_subtract_bit_size_signed(self):
        """
        Perform a bitwise subtract op subtraction.
        """
        ops = [bitwise_operations.bit_subtract(self.five_255_bin, 0, 8, 156, True, aerospike.BIT_OVERFLOW_FAIL, None)]

        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([99] * 1 + [255] * 4)
        assert bins[self.five_255_bin] == expected_result

    def test_bit_subtract_overflow_fail(self):
        """
        Perform a bitwise subtract op that overflows with the BIT_OVERFLOW_FAIL action.
        """
        ops = [bitwise_operations.bit_subtract(self.test_bin_ones, 0, 8, 255, False, aerospike.BIT_OVERFLOW_FAIL, None)]

        with pytest.raises(e.OpNotApplicable):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_subtract_overflow_saturate(self):
        """
        Perform a bitwise subtract op that overflows with the BIT_OVERFLOW_SATURATE action.
        """
        ops = [
            bitwise_operations.bit_subtract(self.test_bin_ones, 0, 8, 255, False, aerospike.BIT_OVERFLOW_SATURATE, None)
        ]

        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([255] * 5)
        assert bins[self.five_255_bin] == expected_result

    def test_bit_subtract_overflow_wrap(self):
        """
        Perform a bitwise subtract op that overflows with the BIT_OVERFLOW_WRAP action.
        """
        ops = [bitwise_operations.bit_subtract(self.five_255_bin, 0, 8, 1, False, aerospike.BIT_OVERFLOW_WRAP, None)]

        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([254] * 1 + [255] * 4)
        assert bins[self.five_255_bin] == expected_result

    def test_bit_xor(self):
        """
        Perform a bitwise xor op.
        """
        value = bytearray([1])
        ops = [bitwise_operations.bit_xor(self.test_bin_ones, 0, 8, 1, value, None)]

        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([0] * 1 + [1] * 4)
        assert bins[self.test_bin_ones] == expected_result

    def test_bit_xor_multiple_bytes(self):
        """
        Perform a bitwise xor op on multiple bytes.
        """
        value = bytearray([8] * 5)
        ops = [bitwise_operations.bit_xor(self.five_255_bin, 0, 40, 5, value, None)]

        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([247] * 5)
        assert bins[self.five_255_bin] == expected_result

    def test_bit_xor_multiple_bytes_value_unchanged(self):
        """
        Perform a bitwise xor op.
        """
        value = bytearray([0])
        ops = [bitwise_operations.bit_xor(self.test_bin_zeroes, 7, 8, 1, value, None)]

        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([0] * 5)
        assert bins[self.test_bin_zeroes] == expected_result

    def test_bit_xor_bit_offset_out_of_range(self):
        """
        Perform a bitwise xor op with bit_offset > bitmap.
        """
        value = bytearray()
        value.append(8)
        ops = [bitwise_operations.bit_xor(self.test_bin_ones, 41, 8, 1, value, None)]

        with pytest.raises(e.OpNotApplicable):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_xor_bit_size_larger_than_value(self):
        """
        Perform a bitwise xor op with bit_size > value.
        """
        value = bytearray()
        value.append(8)
        ops = [bitwise_operations.bit_xor(self.test_bin_ones, 0, 9, 1, value, None)]

        with pytest.raises(e.InvalidRequest):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_xor_bit_size_too_large(self):
        """
        Perform a bitwise xor op with bit_size > bitmap.
        """
        value = bytearray([8] * 6)
        ops = [bitwise_operations.bit_xor(self.test_bin_ones, 0, 41, 6, value, None)]

        with pytest.raises(e.OpNotApplicable):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_xor_bad_bin_name(self):
        """
        Perform a bitwise xor op with a non existent bin.
        """
        value = bytearray([8])
        ops = [bitwise_operations.bit_xor("bad_name", 0, 8, 1, value, None)]

        with pytest.raises(e.BinNotFound):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_xor_bad_arg(self):
        """
        Perform a bitwise xor op with an integer for value.
        """
        value = 1
        ops = [bitwise_operations.bit_xor("bad_name", 0, 8, 1, value, None)]

        with pytest.raises(e.ParamError):
            self.as_connection.operate(self.test_key, ops)

    def test_bit_xor_with_policy(self):
        """
        Perform a bitwise xor op with a policy.
        """
        value = bytearray([0])
        bit_policy = {
            "bit_write_flags": aerospike.BIT_WRITE_UPDATE_ONLY,
        }
        ops = [bitwise_operations.bit_xor(self.test_bin_zeroes, 7, 8, 1, value, bit_policy)]

        self.as_connection.operate(self.test_key, ops)

        _, _, bins = self.as_connection.get(self.test_key)
        expected_result = bytearray([0] * 5)
        assert bins[self.test_bin_zeroes] == expected_result
